import logging

try:
    import cairo
    CAIRO_AVAILABLE = True
except (ImportError, ValueError):
    CAIRO_AVAILABLE = False

try:
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

try:
    import gi
    gi.require_version('Pango', '1.0')
    gi.require_version('PangoCairo', '1.0')
    from gi.repository import Pango
    from gi.repository import PangoCairo
    PANGO_AVAILABLE = True
except (ImportError, ValueError):
    PANGO_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.deps
import openpaperwork_gtk.deps

from .. import BaseDocViewController
from ..... import _


LOGGER = logging.getLogger(__name__)


class BaseDisplayNewDocController(BaseDocViewController):
    def __init__(self, docview, empty_doc_plugin):
        super().__init__(docview)
        self.empty_doc_plugin = empty_doc_plugin
        self.core = empty_doc_plugin.core

    def _switch_controller(self):
        nb_pages = self.core.call_success(
            "doc_get_nb_pages_by_url", self.plugin.active_doc[1]
        )
        if nb_pages is None:
            nb_pages = 0

        if nb_pages > 0 or self.empty_doc_plugin.scanning:
            self.plugin.docview_switch_controller(
                "new_doc",
                lambda docview: NoDisplayController(
                    docview, self.empty_doc_plugin
                )
            )
        else:
            self.plugin.docview_switch_controller(
                "new_doc",
                lambda docview: DisplayNewDocController(
                    docview, self.empty_doc_plugin
                )
            )

    def doc_reload(self):
        self._switch_controller()


class DisplayNewDocController(BaseDisplayNewDocController):
    def __init__(self, docview, plugin):
        super().__init__(docview, plugin)
        self.core = docview.core

    def enter(self):
        self.plugin.overlay.set_visible(True)

    def on_overlay_draw(self, overlay_drawing_area, cairo_ctx):
        if self.empty_doc_plugin.scanning:
            self._switch_controller()
            return

        img = self.empty_doc_plugin.img
        alloc = overlay_drawing_area.get_allocation()
        style = overlay_drawing_area.get_style_context()
        color = style.get_color(Gtk.StateFlags.ACTIVE)

        layout = PangoCairo.create_layout(cairo_ctx)
        layout.set_text(_("Empty"), -1)  # text must be short
        txt_size = layout.get_size()
        if 0 in txt_size:
            return

        zoom = img.get_height() / txt_size[1]
        scaled_txt_width = txt_size[0] * zoom

        cairo_ctx.save()
        try:
            cairo_ctx.set_source_rgba(color.red, color.green, color.blue, 0.33)
            cairo_ctx.mask_surface(
                self.empty_doc_plugin.img,
                ((alloc.width - img.get_width() - scaled_txt_width) / 2),
                ((alloc.height - img.get_height()) / 2),
            )
            cairo_ctx.fill()
        finally:
            cairo_ctx.restore()

        cairo_ctx.save()
        try:
            cairo_ctx.set_source_rgba(color.red, color.green, color.blue, 0.33)
            cairo_ctx.translate(
                ((alloc.width - img.get_width() - scaled_txt_width) / 2) +
                img.get_width(),
                ((alloc.height - img.get_height()) / 2),
            )
            cairo_ctx.scale(zoom * Pango.SCALE, zoom * Pango.SCALE)
            PangoCairo.update_layout(cairo_ctx, layout)
            PangoCairo.show_layout(cairo_ctx, layout)
        finally:
            cairo_ctx.restore()


class NoDisplayController(BaseDisplayNewDocController):
    def enter(self):
        self.plugin.overlay.set_visible(False)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.scanning = False

    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_docview_controller',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_gtk.fs.gio'],
            },
            {
                'interface': 'gtk_docview',
                'defaults': ['paperwork_gtk.mainwindow.docview'],
            },
            {
                'interface': 'resources',
                'defaults': ['openpaperwork_core.resources.setuptools'],
            },
        ]

    def init(self, core):
        super().init(core)

        if not CAIRO_AVAILABLE:
            # chkdeps() must still be callable
            return

        file_path_ctx = self.core.call_success(
            "resources_get_file",
            "paperwork_gtk.mainwindow.docview.controllers.empty_doc",
            "empty_doc.png"
        )
        with file_path_ctx as file_path:
            self.img = cairo.ImageSurface.create_from_png(file_path)

    def chkdeps(self, out: dict):
        if not CAIRO_AVAILABLE:
            out['cairo'].update(openpaperwork_core.deps.CAIRO)
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)
        if not PANGO_AVAILABLE:
            out['pango'].update(openpaperwork_core.deps.PANGO)

    def on_scan2doc_start(self, *args, **kwargs):
        self.scanning = True

    def on_scan2doc_end(self, *args, **kwargs):
        self.scanning = False

    def gtk_docview_get_controllers(self, out: dict, docview):
        nb_pages = self.core.call_success(
            "doc_get_nb_pages_by_url", docview.active_doc[1]
        )
        if nb_pages is None:
            nb_pages = 0

        out['new_doc'] = (
            DisplayNewDocController(docview, self)
            if nb_pages <= 0 else
            NoDisplayController(docview, self)
        )



import logging

import openpaperwork_core
import openpaperwork_core.promise


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.widget_tree = None
        self.default_action = None
        self.default_action_args = None
        self.active_doc_id = None
        self.active_doc_url = None
        self.busy = 0

    def get_interfaces(self):
        return [
            'doc_open',
            'gtk_scan_buttons',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'gtk_docview',
                'defaults': ['paperwork_gtk.mainwindow.docview'],
            },
            {
                'interface': 'new_doc',
                'defaults': ['paperwork_gtk.new_doc'],
            },
            {
                'interface': 'screenshot',
                'defaults': ['openpaperwork_gtk.screenshots'],
            },
        ]

    def init(self, core):
        super().init(core)

        self.widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.mainwindow.docview.pageadd", "buttons.glade"
        )
        if self.widget_tree is None:
            # init must still work so 'chkdeps' is still available
            LOGGER.error("Failed to load widget tree")
            return

        self.widget_tree.get_object("pageadd_button").connect(
            "clicked", self._on_clicked
        )

        headerbar = self.core.call_success("docview_get_headerbar")
        headerbar.pack_start(self.widget_tree.get_object("pageadd_buttons"))

    def pageadd_buttons_set_source_popover(self, selector):
        self.widget_tree.get_object("pageadd_switch").set_popover(selector)

    def _update_sensitivity(self):
        sensitive = (
            self.default_action is not None and self.busy <= 0
        )
        button = self.widget_tree.get_object("pageadd_button")
        button.set_sensitive(sensitive)
        self.core.call_all("on_widget_busyness_changed", button, sensitive)

    def pageadd_set_default_action(self, txt, callback, *args):
        self.default_action = callback
        self.default_action_args = args
        self.widget_tree.get_object("pageadd_button").set_label(txt)
        self._update_sensitivity()

    def pageadd_busy_add(self):
        self.busy += 1
        self._update_sensitivity()

    def pageadd_busy_remove(self):
        self.busy -= 1
        self._update_sensitivity()

    def _on_clicked(self, widget):
        self.default_action(
            self.active_doc_id, self.active_doc_url,
            *self.default_action_args
        )

    def doc_open(self, doc_id, doc_url):
        self.active_doc_id = doc_id
        self.active_doc_url = doc_url
        self._update_sensitivity()

    def doc_close(self):
        self.active_doc_id = None
        self.active_doc_url = None
        self._update_sensitivity()

    def on_scan_feed_start(self, scan_id):
        self.pageadd_busy_add()

    def on_scan_feed_end(self, scan_id):
        self.pageadd_busy_remove()

    def screenshot_snap_all_doc_widgets(self, out_dir):
        self.core.call_success(
            "screenshot_snap_widget",
            self.widget_tree.get_object("pageadd_buttons"),
            self.core.call_success("fs_join", out_dir, "page_add.png")
        )



import logging

try:
    from gi.repository import Gio
    GLIB_AVAILABLE = True
except (ImportError, ValueError):
    GLIB_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.deps


LOGGER = logging.getLogger(__name__)
ACTION_NAME = "doc_open_external"


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.active_doc = None
        self.active_windows = []

    def get_interfaces(self):
        return [
            'action',
            'action_doc',
            'action_doc_open_external',
            'chkdeps',
            'doc_open',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'app_actions',
                'defaults': ['paperwork_gtk.mainwindow.window'],
            },
            {
                'interface': 'external_apps',
                'defaults': [
                    'openpaperwork_core.external_apps.dbus',
                    'openpaperwork_core.external_apps.windows',
                    'openpaperwork_core.external_apps.xdg',
                ],
            },
        ]

    def init(self, core):
        super().init(core)
        if not GLIB_AVAILABLE:
            return
        action = Gio.SimpleAction.new(ACTION_NAME, None)
        action.connect("activate", self._open_external)
        self.core.call_all("app_actions_add", action)

    def chkdeps(self, out: dict):
        if not GLIB_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)

    def doc_open(self, doc_id, doc_url):
        self.active_doc = (doc_id, doc_url)

    def doc_close(self):
        self.active_doc = None

    def _open_external(self, action, parameter):
        assert self.active_doc is not None
        (doc_id, doc_url) = self.active_doc
        self.core.call_success("external_app_open_folder", doc_url)



import collections
import logging

try:
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

import openpaperwork_core
import openpaperwork_gtk.deps

import paperwork_backend.pageedit


LOGGER = logging.getLogger(__name__)
MODIFIER_ICONS = collections.defaultdict(
    lambda: "document-properties",
    {
        "color_equalization": (
            "paperwork_gtk.mainwindow.pageeditor", "magic_colors.png"
        ),
        "crop": "edit-cut-symbolic",
        "rotate_clockwise": "object-rotate-left-symbolic",
        "rotate_counterclockwise": "object-rotate-right-symbolic",
    }
)


class GtkPageEditorUI(paperwork_backend.pageedit.AbstractPageEditorUI):
    CAPABILITIES = (
        paperwork_backend.pageedit.AbstractPageEditorUI.CAPABILITY_SHOW_FRAME
    )

    def __init__(self, plugin):
        super().__init__()
        self.plugin = plugin
        self.core = plugin.core
        self.editor = None

        self.modifiers_to_toggles = {}
        self.buttons_to_modifiers = {}

        self.pil_img = None
        self.surface_img = None

        self.size_allocated = None
        self.allocate_handler_id = None

        self.draw_handler_id = self.plugin.widget_tree.get_object(
            "pageeditor_img"
        ).connect("draw", self.draw)

        self.zoom = self.plugin.widget_tree.get_object(
            "pageeditor_zoom_adjustment"
        )
        self.zoom.connect("value-changed", self._on_zoom_changed)

    def add_modifier_toggles(self):
        assert self.editor is not None

        toolbox = self.plugin.widget_tree.get_object("pageeditor_tools")
        zoom_box = self.plugin.widget_tree.get_object("pageeditor_zoom_box")
        for widget in list(toolbox.get_children()):
            if widget == zoom_box:
                continue
            toolbox.remove(widget)

        for modifier in self.editor.get_modifiers():
            glade = (
                "modifier_toggle.glade"
                if modifier['togglable'] else
                "modifier.glade"
            )
            modifier_widget_tree = self.core.call_success(
                "gtk_load_widget_tree",
                "paperwork_gtk.mainwindow.pageeditor", glade
            )

            button = modifier_widget_tree.get_object("pageeditor_modifier")
            button.set_tooltip_text(modifier['name'])
            if modifier['togglable']:
                button.set_active(modifier['enabled'])
                button.connect("toggled", self._on_modifier_changed)
                self.modifiers_to_toggles[modifier['id']] = button
            else:
                button.connect("clicked", self._on_modifier_changed)

            img = modifier_widget_tree.get_object("pageeditor_modifier_img")
            icon = MODIFIER_ICONS[modifier['id']]
            if isinstance(icon, tuple):
                icon = self.core.call_success("gtk_load_pixbuf", *icon)
                img.set_from_pixbuf(icon)
            else:
                img.set_from_icon_name(icon, Gtk.IconSize.LARGE_TOOLBAR)

            toolbox.pack_start(button, expand=False, fill=True, padding=0)

            self.buttons_to_modifiers[button] = modifier['id']

        if self.size_allocated is None:
            self.allocate_handler_id = self.plugin.widget_tree.get_object(
                "pageeditor_scroll"
            ).connect("size-allocate", self._on_size_allocate)
        else:
            self.set_default_zoom()

        self.plugin.widget_tree.get_object(
            "pageeditor_img"
        ).connect("realize", self.refresh)

    def _on_modifier_changed(self, button):
        modifier = self.buttons_to_modifiers[button]
        promise = openpaperwork_core.promise.Promise(
            self.core, self.core.call_all, args=("on_busy",)
        )
        promise = promise.then(lambda *args, **kwargs: None)
        promise = promise.then(self.editor.on_modifier_selected(modifier))
        promise.then(self.refresh)
        promise.then(self.core.call_all, "on_idle")
        promise.schedule()

    def _on_size_allocate(self, *args, **kwargs):
        self.plugin.widget_tree.get_object(
            "pageeditor_scroll"
        ).disconnect(self.allocate_handler_id)
        self.allocate_handler_id = None

        self.set_default_zoom()

    def set_default_zoom(self, *args, **kwargs):
        allocation = self.plugin.widget_tree.get_object(
            "pageeditor_scroll"
        ).get_allocation()
        img_size = self.pil_img.size

        zoom = min(
            (allocation.width - 20) / img_size[0],
            (allocation.height - 20) / img_size[1]
        )
        LOGGER.info(
            "Allocation: %dx%d ; Image: %dx%d ==> Setting zoom at %f",
            allocation.width, allocation.height, img_size[0], img_size[1],
            zoom
        )
        self.zoom.set_value(zoom)
        self.refresh()

    def _get_scaled_image_size(self):
        img_size = self.pil_img.size
        zoom = self.zoom.get_value()
        return (img_size[0] * zoom, img_size[1] * zoom)

    def refresh(self, *args, **kwargs):
        widget_size = self._get_scaled_image_size()
        img = self.plugin.widget_tree.get_object("pageeditor_img")
        img.set_size_request(widget_size[0], widget_size[1])

        # WORKAROUND(JFlesch): I shouldn't have to use 'mainloop_schedule'
        # here. But somehow, I do have to use it :/
        self.core.call_all("mainloop_schedule", img.queue_resize)

    def _on_zoom_changed(self, adj=None):
        self.refresh()

    def set_modifier_state(self, modifier_id, enabled):
        super().set_modifier_state(modifier_id, enabled)
        LOGGER.info("Modifier %s: %s", modifier_id, enabled)

    def show_preview(self, img):
        super().show_preview(img)
        need_rezoom = (self.pil_img is None)

        self.pil_img = img
        self.surface_img = self.core.call_success(
            "pillow_to_surface", self.pil_img
        )

        if need_rezoom:
            self.set_default_zoom()

        self.refresh()
        LOGGER.info("Preview refreshed (%s)", img.size)

    def show_frame_selector(self):
        super().show_frame_selector()
        if self.pil_img is None:
            return
        img = self.plugin.widget_tree.get_object("pageeditor_img")
        self.core.call_all("draw_frame_stop", img)

        LOGGER.info(
            "show_frame_selector() (img size: %s ; frame: %s)",
            self.pil_img.size, self.editor.frame.get()
        )
        self.core.call_all(
            "draw_frame_start", img, self.pil_img.size,
            self.editor.frame.get, self.editor.frame.set
        )

    def hide_frame_selector(self):
        super().hide_frame_selector()
        LOGGER.info("hide_frame_selector()")
        img = self.plugin.widget_tree.get_object("pageeditor_img")
        self.core.call_all("draw_frame_stop", img)

    def on_edit_end(self, doc_url, page_idx):
        super().on_edit_end(doc_url, page_idx)
        self.plugin._on_edit_end(doc_url, page_idx)
        self._disconnect_draw()
        self.surface_img = None

    def _disconnect_draw(self):
        if self.draw_handler_id is None:
            return
        self.plugin.widget_tree.get_object(
            "pageeditor_img"
        ).disconnect(self.draw_handler_id)
        self.draw_handler_id = None

    def cancel(self):
        self.editor.on_cancel()
        self._disconnect_draw()

    def save(self):
        self.core.call_all("on_busy")
        self.plugin.widget_tree.get_object("pageeditor_back").set_sensitive(
            False
        )
        promise = openpaperwork_core.promise.Promise(self.core)
        promise = promise.then(lambda *args, **kwargs: None)
        promise = promise.then(self.editor.on_save())
        promise = promise.then(self.core.call_all, "on_idle")
        promise = promise.then(lambda *args, **kwargs: None)
        promise = promise.then(self.core.call_success(
            "transaction_simple_promise",
            [("upd", self.plugin.active_doc[0])]
        ))
        self.core.call_success("transaction_schedule", promise)

    def draw(self, widget, cairo_ctx):
        if self.surface_img is None:
            return
        cairo_ctx.save()
        try:
            zoom = self.zoom.get_value()
            cairo_ctx.scale(zoom, zoom)
            cairo_ctx.set_source_surface(self.surface_img.surface)
            cairo_ctx.paint()
        finally:
            cairo_ctx.restore()


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.ui = None
        self.active_doc = None

    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_page_editor',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'gtk_drawer_frame',
                'defaults': ['paperwork_gtk.drawer.frame'],
            },
            {
                'interface': 'gtk_mainwindow',
                'defaults': ['paperwork_gtk.mainwindow.window'],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'gtk_zoomable',
                'defaults': ['paperwork_gtk.gesture.zoom'],
            },
            {
                'interface': 'page_editor',
                'defaults': ['paperwork_backend.pageedit.pageeditor'],
            },
            {
                'interface': 'pillow_to_surface',
                'defaults': ['paperwork_backend.cairo.pillow'],
            },
            {
                'interface': 'resources',
                'defaults': ['openpaperwork_core.resources.setuptools'],
            },
            {
                'interface': 'transaction_manager',
                'defaults': ['paperwork_backend.sync'],
            },
        ]

    def init(self, core):
        super().init(core)

        self.widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.mainwindow.pageeditor", "pageeditor.glade"
        )
        if self.widget_tree is None:
            # init must still work so 'chkdeps' is still available
            LOGGER.error("Failed to load widget tree")
            return

        self.widget_tree.get_object("pageeditor_cancel").connect(
            "clicked", self._on_cancel
        )
        self.widget_tree.get_object("pageeditor_back").connect(
            "clicked", self._on_apply
        )

        self.core.call_all(
            "on_zoomable_widget_new",
            self.widget_tree.get_object("pageeditor_scroll"),
            self.widget_tree.get_object("pageeditor_zoom_adjustment"),
        )

        self.core.call_all(
            "mainwindow_add", "right", "pageeditor", prio=0,
            header=self.widget_tree.get_object("pageeditor_header"),
            body=self.widget_tree.get_object("pageeditor_body")
        )

    def chkdeps(self, out: dict):
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)

    def doc_open(self, doc_id, doc_url):
        self._on_cancel(None)
        self.active_doc = (doc_id, doc_url)

    def gtk_open_page_editor(self, doc_id, doc_url, page_idx):
        LOGGER.info("Opening page editor on %s %d", doc_id, page_idx)
        self.active_doc = (doc_id, doc_url)

        page_url = self.core.call_success(
            "page_get_img_url", doc_url, page_idx
        )
        if page_url is None:
            LOGGER.error(
                "Can't open page editor: Failed to get page url (%s, p%d)",
                doc_id, page_idx
            )
            return

        self.ui = GtkPageEditorUI(self)
        self.widget_tree.get_object("pageeditor_back").set_sensitive(True)

        promise = openpaperwork_core.promise.Promise(
            self.core, self.core.call_all, args=("on_busy",)
        )
        promise = promise.then(lambda *args, **kwargs: None)
        promise = promise.then(
            self.core.call_success, "page_editor_get", doc_url, page_idx,
            self.ui
        )
        promise = promise.then(
            self._show_page_editor, self.ui, doc_id, doc_url, page_idx
        )
        promise = promise.then(lambda *args, **kwargs: None)
        promise = promise.then(self.core.call_all, "on_idle")
        promise.schedule()

    def _on_cancel(self, button):
        if self.ui is None:
            return
        self.ui.cancel()
        self.ui = None

    def _on_apply(self, button):
        if self.ui is None:
            return
        self.ui.save()

    def _on_edit_end(self, doc_url, page_idx):
        self.core.call_success(
            "mainwindow_show", side="right", name="docview"
        )
        self.core.call_all("doc_reload", *self.active_doc)
        self.core.call_success(
            "mainloop_schedule", self.core.call_all, "doc_goto_page", page_idx
        )
        self.ui = None

    def _show_page_editor(self, page_editor, ui, doc_id, doc_url, page_idx):
        self.ui.editor = page_editor
        self.ui.add_modifier_toggles()
        self.core.call_all("mainwindow_show", "right", "pageeditor")



import logging
import re

import openpaperwork_core


LOGGER = logging.getLogger(__name__)

MIN_WORD_LENGTH = 3
SPLIT = r"\W+"


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 100

    # those are keywords used in python-whoosh query syntax. There is no
    # point in highlighting them
    IGNORE_LIST = {"and", "or"}

    def __init__(self):
        super().__init__()
        self.re_split = re.compile(SPLIT)
        self.keywords = set()

    def get_interfaces(self):
        return [
            'gtk_pageview_boxes_listener',
            'search_listener',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'gtk_pageview',
                'defaults': ['paperwork_gtk.mainwindow.docview.pageview'],
            },
            {
                'interface': 'gtk_pageview_boxes',
                'defaults': [
                    'paperwork_gtk.mainwindow.docview.pageview.boxes'
                ],
            },
            {
                'interface': 'i18n',
                'defaults': ['openpaperwork_core.i18n.python'],
            },
        ]

    def on_search_start(self, query):
        query = self.re_split.split(
            self.core.call_success("i18n_strip_accents", query)
        )
        self.keywords = {
            keyword.lower() for keyword in query
            if len(keyword) >= MIN_WORD_LENGTH
        }
        self.core.call_all("pageview_refresh_all")

    def on_search_results(self, query, docs):
        pass

    def on_page_draw(self, cairo_ctx, page):
        if len(self.keywords) <= 0:
            return

        boxes = self.core.call_success(
            "pageview_get_boxes_by_id", page.doc_id, page.page_idx
        )
        if boxes is None:
            return

        for line_box in boxes:
            for word_box in line_box.word_boxes:
                word = self.core.call_success(
                    "i18n_strip_accents", word_box.content
                )
                word = word.strip()
                if word.lower() in self.IGNORE_LIST:
                    continue
                for w in self.re_split.split(word):
                    w = w.lower()
                    if w not in self.keywords:
                        continue
                    self.core.call_all(
                        "page_draw_box",
                        cairo_ctx, page, word_box.position,
                        (0.0, 1.0, 0.0), border_width=2
                    )



import datetime
import logging
import re

try:
    from gi.repository import GLib
    from gi.repository import GObject
    GLIB_AVAILABLE = True
except (ImportError, ValueError):
    GLIB_AVAILABLE = False

try:
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

import openpaperwork_core
import openpaperwork_gtk.deps

from ... import _


LOGGER = logging.getLogger(__name__)


def strip_quotes(txt):
    if txt[0] == u'"' and txt[-1] == u'"':
        return txt[1:-1]
    if txt[0] == u'\'' and txt[-1] == u'\'':
        return txt[1:-1]
    return txt


class SearchElement(object):
    def __init__(self, dialog, widget):
        self.dialog = dialog
        self.widget = widget
        widget.set_hexpand(True)
        widget.show_all()

    def get_widget(self):
        return self.widget

    def get_search_string(self):
        assert False

    @staticmethod
    def get_from_search(dialog, text):
        assert False

    @staticmethod
    def get_name():
        assert False

    def on_model_updated(self):
        pass


class SearchElementText(SearchElement):
    """This is the keyword search term text field"""
    def __init__(self, dialog):
        super(SearchElementText, self).__init__(dialog, Gtk.Entry())

    def get_search_string(self):
        txt = self.widget.get_text()
        txt = txt.replace('"', '\\"')
        return '"%s"' % txt

    @staticmethod
    def get_from_search(dialog, text):
        text = strip_quotes(text)
        element = SearchElementText(dialog)
        element.widget.set_text(text)
        return element

    @staticmethod
    def get_name():
        return _("Keyword(s)")

    def __str__(self):
        return ("Text: [%s]" % self.widget.get_text())


class SearchElementLabel(SearchElement):
    def __init__(self, dialog):
        super().__init__(dialog, Gtk.ComboBoxText())
        self.on_model_updated()
        self.widget.set_active(0)

    def on_model_updated(self):
        labels = set()
        self.dialog.core.call_all("labels_get_all", labels)
        labels = [label[0] for label in labels]
        labels = self.dialog.core.call_success("i18n_sort", labels)

        store = Gtk.ListStore.new([GObject.TYPE_STRING])
        if len(labels) <= 0:
            # happens temporarily when Paperwork starts
            store.append([_("No labels")])
            self.widget.set_sensitive(False)
        else:
            for label in labels:
                store.append([label])
            self.widget.set_sensitive(True)
        self.widget.set_model(store)

    def get_search_string(self):
        active_idx = self.get_widget().get_active()
        if active_idx < 0:
            return ""
        model = self.get_widget().get_model()
        txt = model[active_idx][0]
        txt = txt.replace('"', '\\"')
        return 'label:"%s"' % txt

    @staticmethod
    def get_from_search(dialog, text):
        if not text.startswith(u"label:"):
            return None

        text = text[len(u"label:"):]
        text = strip_quotes(text)

        element = SearchElementLabel(dialog)

        active_idx = -1
        idx = 0
        for line in element.get_widget().get_model():
            value = line[0]
            if value == text:
                active_idx = idx
            idx += 1
        element.get_widget().set_active(active_idx)

        return element

    @staticmethod
    def get_name():
        return _("Label")

    def __str__(self):
        return ("Label: [%d]" % self.get_widget().get_active())


class SearchElementDate(SearchElement):
    """Search entry using a time span"""
    def __init__(self, dialog):
        box = Gtk.Box()
        box.set_spacing(10)

        label = Gtk.Label.new(_("From:"))
        box.add(label)

        self.start_date = self._make_date_widget()
        box.add(self.start_date)

        label = Gtk.Label.new(_("to:"))
        box.add(label)

        self.end_date = self._make_date_widget()
        box.add(self.end_date)
        super(SearchElementDate, self).__init__(dialog, box)

        self.calendar_popover = dialog.widget_tree.get_object(
            "calendar_popover"
        )
        self.calendar = dialog.widget_tree.get_object("calendar_calendar")

        self.current_entry = None
        self.calendar.connect(
            "day-selected-double-click",
            lambda _: GLib.idle_add(self._close_calendar)
        )

    def _make_date_widget(self):
        entry = Gtk.Entry()
        entry.set_text("")
        entry.set_property("secondary_icon_sensitive", True)
        entry.set_property("secondary_icon_name", "x-office-calendar-symbolic")
        entry.connect(
            "icon-release",
            lambda entry, icon, event:
            GLib.idle_add(self._open_calendar, entry)
        )
        return entry

    @staticmethod
    def _parse_date(txt):
        txt = txt.strip()
        if txt == u"":
            dt = datetime.datetime.today()
        else:
            try:
                dt = datetime.datetime.strptime(txt, "%Y%m%d")
            except ValueError:
                LOGGER.warning(
                    "Failed to parse [%s]. Will use today date", txt
                )
                dt = datetime.datetime.today()
        return (dt.year, dt.month, dt.day)

    @staticmethod
    def _format_date(date):
        return "%04d%02d%02d" % (date[0], date[1], date[2])

    def _open_calendar(self, entry):
        self.calendar_popover.set_relative_to(entry)
        date = self._parse_date(entry.get_text())
        self.calendar.select_month(date[1] - 1, date[0])
        self.calendar.select_day(date[2])
        self.calendar_popover.show_all()
        self.current_entry = entry

    def _close_calendar(self):
        date = self.calendar.get_date()
        date = datetime.datetime(year=date[0], month=date[1] + 1, day=date[2])
        date = self._format_date((date.year, date.month, date.day))
        self.current_entry.set_text(date)
        self.calendar_popover.set_visible(False)

    def get_search_string(self):
        start_date = self._parse_date(self.start_date.get_text())
        end_date = self._parse_date(self.end_date.get_text())
        if end_date < start_date:
            tmp_date = start_date
            start_date = end_date
            end_date = tmp_date
        if start_date == end_date:
            return (
                "date:%04d%02d%02d"
                % (start_date[0], start_date[1], start_date[2])
            )
        return (
            'date:[%04d%02d%02d to %04d%02d%02d]' % (
                start_date[0], start_date[1], start_date[2],
                end_date[0], end_date[1], end_date[2]
            )
        )

    @staticmethod
    def get_from_search(dialog, txt):
        if not txt.startswith(u"date:"):
            return None

        txt = txt[len(u"date:"):]
        txt = strip_quotes(txt)

        if txt[0] == "[" and txt[-1] == "]":
            txt = txt[1:-1]
        if " to " in txt:
            txt = txt.split(" to ", 1)
        else:
            txt = [txt, txt]

        dates = [
            SearchElementDate._parse_date(date)
            for date in txt
        ]

        se = SearchElementDate(dialog)
        se.start_date.set_text(se._format_date(dates[0]))
        se.end_date.set_text(se._format_date(dates[1]))
        return se

    @staticmethod
    def get_name():
        return _("Date")

    def __str__(self):
        return (
            "Date: [%s] - [%s]"
            % (self.start_date.get_text(), self.end_date.get_text())
        )


class SearchLine(object):
    SELECT_ORDER = [
        SearchElementText,
        SearchElementLabel,
        SearchElementDate,
    ]
    TXT_EVAL_ORDER = [
        SearchElementDate,
        SearchElementLabel,
        SearchElementText,
    ]

    def __init__(self, dialog, has_operator, has_remove_button):
        LOGGER.info("Search line instantiated")

        self.dialog = dialog
        self.line = []

        if has_operator:
            model = Gtk.ListStore.new([
                GObject.TYPE_STRING,
                GObject.TYPE_STRING,
            ])
            model.append([_("and"), "AND"])
            model.append([_("or"), "OR"])
            self.combobox_operator = Gtk.ComboBoxText.new()
            self.combobox_operator.set_model(model)
            self.combobox_operator.set_size_request(75, -1)
            self.combobox_operator.set_active(0)
            self.line.append(self.combobox_operator)
        else:
            self.combobox_operator = None
            placeholder = Gtk.Label.new("")
            placeholder.set_size_request(75, -1)
            self.line.append(placeholder)

        model = Gtk.ListStore.new([
            GObject.TYPE_STRING,
            GObject.TYPE_STRING,
        ])
        model.append(["", ""])
        model.append([_("not"), "NOT"])
        self.combobox_not = Gtk.ComboBoxText.new()
        self.combobox_not.set_model(model)
        self.combobox_not.set_size_request(75, -1)
        self.combobox_not.set_active(0)
        self.line.append(self.combobox_not)

        model = Gtk.ListStore.new([
            GObject.TYPE_STRING,
            GObject.TYPE_PYOBJECT,
        ])
        for element in self.SELECT_ORDER:
            model.append([
                element.get_name(),
                element
            ])

        self.combobox_type = Gtk.ComboBoxText.new()
        self.combobox_type.set_model(model)

        self.placeholder = Gtk.Label.new("")
        self.placeholder.set_hexpand(True)

        self.element = None
        self.remove_button = Gtk.Button.new_with_label(_("Remove"))

        self.line.append(self.combobox_type)
        self.line.append(self.placeholder)
        if has_remove_button:
            self.line.append(self.remove_button)
        else:
            self.line.append(Gtk.Label.new(""))

        self.combobox_type.set_active(0)

        self.change_element()

    def connect_signals(self):
        self.combobox_type.connect(
            "changed", lambda w: GLib.idle_add(self.change_element)
        )
        self.remove_button.connect(
            "clicked",
            lambda x: GLib.idle_add(
                self.dialog.remove_element,
                self
            )
        )

    @staticmethod
    def _select_value(combobox, value):
        if not combobox:
            return
        active_idx = 0
        model = combobox.get_model()
        for line in model:
            if line[1] == value:
                LOGGER.info("Element %d selected", active_idx)
                combobox.set_active(active_idx)
                return
            active_idx += 1
        assert False

    def select_operator(self, operator):
        self._select_value(self.combobox_operator, operator.upper())

    def select_not(self, not_value):
        self._select_value(self.combobox_not, not_value)

    def select_element_type(self, et):
        self._select_value(self.combobox_type, et)

    def change_element(self):
        LOGGER.info("Element changed")
        active_idx = self.combobox_type.get_active()
        if (active_idx < 0):
            return
        element_class = self.combobox_type.get_model()[active_idx][1]
        element = element_class(self.dialog)
        self.set_element(element)

    def set_element(self, element):
        LOGGER.info("Set element: %s", str(element))
        if self.placeholder:
            self.line.remove(self.placeholder)
            self.placeholder = None
        if self.element:
            self.line.remove(self.element.get_widget())
            self.element = None
        self.line.insert(3, element.get_widget())
        self.element = element
        self.dialog.rebuild()

    def get_widgets(self):
        return self.line

    @staticmethod
    def _get_combobox_value(combobox):
        active_idx = combobox.get_active()
        if (active_idx < 0):
            return ""
        value = combobox.get_model()[active_idx][1]
        return value

    def get_operator(self):
        if not self.combobox_operator:
            return u""
        return self._get_combobox_value(self.combobox_operator)

    def get_not(self):
        return self._get_combobox_value(self.combobox_not)

    def get_search_string(self):
        if self.element is None:
            return ""
        return self.element.get_search_string()

    @staticmethod
    def get_from_search(dialog, next_operator, not_value, search_txt):
        for se_class in SearchLine.TXT_EVAL_ORDER:
            se = se_class.get_from_search(dialog, search_txt)
            if not se:
                continue
            sl = SearchLine(
                dialog,
                next_operator is not None,
                next_operator is not None
            )
            if next_operator:
                sl.select_operator(next_operator)
            sl.select_element_type(se_class)
            sl.select_not(not_value)
            sl.set_element(se)
            sl.connect_signals()
            LOGGER.info(
                "Loaded from search: %s --> %s", search_txt, str(se)
            )
            return sl
        assert False

    def on_model_updated(self):
        if self.element:
            self.element.on_model_updated()


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.widget_tree = None
        self.windows = []
        self.search_element_box = None
        self.search_elements = []
        self.dialog = None

    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_advanced_search_dialog',
            'gtk_window_listener',
            'screenshot_provider',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'doc_labels',
                'defaults': ['paperwork_backend.model.labels'],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'i18n',
                'defaults': ['openpaperwork_core.i18n.python'],
            },
            {
                'interface': 'screenshot',
                'defaults': ['openpaperwork_gtk.screenshots'],
            },
        ]

    def on_gtk_window_opened(self, window):
        self.windows.append(window)

    def on_gtk_window_closed(self, window):
        self.windows.remove(window)

    def chkdeps(self, out: dict):
        if not GLIB_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)

    def gtk_open_advanced_search_dialog(self):
        if self.dialog is not None:
            LOGGER.warning("Advanced search dialog already opened")
            return

        self.widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.mainwindow.search", "advanced.glade"
        )
        if self.widget_tree is None:
            # init must still work so 'chkdeps' is still available
            LOGGER.error("Failed to load widget tree")
            return

        self.dialog = self.widget_tree.get_object("search_dialog")
        self.dialog.set_transient_for(self.windows[-1])

        keywords = self.core.call_success("search_get")
        keywords = keywords.strip()
        keywords = re.findall(
            r'(?:\[.*\]|(?:[^\s"]|"(?:\\.|[^"])*"))+', keywords
        )

        self.search_element_box = self.widget_tree.get_object(
            "box_search_elements"
        )
        self.search_elements = []

        add_button = self.widget_tree.get_object("button_add")
        add_button.connect(
            "clicked", lambda w: GLib.idle_add(self.add_element)
        )

        if keywords == []:
            LOGGER.info("Starting from an empty search")
            self.add_element()
        else:
            LOGGER.info("Current search: %s", keywords)

            next_operator = None
            not_value = u""
            for keyword in keywords:
                if keyword.upper() == u"AND":
                    next_operator = u"AND"
                    continue
                elif keyword.upper() == u"OR":
                    next_operator = u"OR"
                    continue
                elif keyword.upper() == u"NOT":
                    not_value = u"NOT"
                    continue

                LOGGER.info("Instantiating line for [%s]", keyword)
                sl = SearchLine.get_from_search(
                    self, next_operator, not_value, keyword
                )
                self.add_element(sl)

                next_operator = u"AND"
                not_value = u""

        self.dialog.connect("response", self._on_response)
        self.dialog.show_all()

    def _on_response(self, dialog, response_id):
        if (response_id == Gtk.ResponseType.ACCEPT or
                response_id == Gtk.ResponseType.OK or
                response_id == Gtk.ResponseType.YES or
                response_id == Gtk.ResponseType.APPLY):
            search = self._get_search_string()
            self.core.call_all("search_set", search)
        self.gtk_close_advanced_search_dialog()

    def gtk_close_advanced_search_dialog(self):
        self.dialog.destroy()
        self.widget_tree = None
        self.dialog = None

    def add_element(self, sl=None):
        if sl is None:
            sl = SearchLine(
                self,
                len(self.search_elements) > 0,
                len(self.search_elements) > 0
            )
            sl.connect_signals()
        self.search_elements.append(sl)
        self.rebuild()

    def rebuild(self):
        # purge
        for child in self.search_element_box.get_children():
            self.search_element_box.remove(child)

        # rebuild
        for (line, sl) in enumerate(self.search_elements):
            for (pos, widget) in enumerate(sl.get_widgets()):
                self.search_element_box.attach(
                    widget,
                    pos, line,
                    1, 1
                )
        self.search_element_box.show_all()

    def remove_element(self, sl):
        self.search_elements.remove(sl)
        self.rebuild()

    def _get_search_string(self):
        """concat all our search terms into a single string"""
        out = ""
        for element in self.search_elements:
            # Add AND/OR
            oper = element.get_operator()
            out += " %s " % oper
            not_value = element.get_not()
            if not_value:
                out += "%s " % not_value
            out += element.get_search_string()
        out = out.strip()
        LOGGER.info("Search: [%s]", out)
        return out

    def screenshot_snap_all_doc_widgets(self, out_dir):
        if self.dialog is None:
            return
        self.core.call_success(
            "screenshot_snap_widget",
            self.dialog,
            self.core.call_success(
                "fs_join", out_dir, "advanced_search.png"
            ),
        )

    def on_label_loading_end(self):
        for element in self.search_elements:
            element.on_model_updated()



import logging

import pillowfight


from . import (
    AbstractExportPipePlugin,
    AbstractSimpleTransformExportPipe
)
from .. import _


LOGGER = logging.getLogger(__name__)


class UnpaperExportPipe(AbstractSimpleTransformExportPipe):
    def __init__(self, core):
        super().__init__(core, "unpaper")

    def transform(self, img):
        # Unpaper algorithms in Unpaper's order
        img = pillowfight.unpaper_blackfilter(img)
        img = pillowfight.unpaper_noisefilter(img)
        img = pillowfight.unpaper_blurfilter(img)
        img = pillowfight.unpaper_masks(img)
        img = pillowfight.unpaper_grayfilter(img)
        img = pillowfight.unpaper_border(img)
        return img

    def __str__(self):
        return _("Soft simplification")


class SwtExportPipe(AbstractSimpleTransformExportPipe):
    def __init__(self, core, swt_output_type):
        super().__init__(
            core,
            "swt_soft"
            if swt_output_type == pillowfight.SWT_OUTPUT_ORIGINAL_BOXES
            else "swt_hard"
        )
        self.swt_output_type = swt_output_type

    def transform(self, pil_img):
        pil_img = pil_img.convert("L")
        return pillowfight.swt(pil_img, output_type=self.swt_output_type)

    def __str__(self):
        if self.swt_output_type == pillowfight.SWT_OUTPUT_ORIGINAL_BOXES:
            return _("Hard simplification")
        else:
            return _("Extreme simplification")


class Plugin(AbstractExportPipePlugin):
    def init(self, core):
        super().init(core)
        self.pipes = [
            UnpaperExportPipe(core),
            SwtExportPipe(core, pillowfight.SWT_OUTPUT_ORIGINAL_BOXES),
            SwtExportPipe(core, pillowfight.SWT_OUTPUT_BW_TEXT),
        ]

    def get_deps(self):
        return [
            {
                'interface': 'mainloop',
                'defaults': 'openpaperwork_gtk.mainloop.glib',
            },
        ]



import logging
import heapq
import threading
import traceback

from .. import PluginBase


LOGGER = logging.getLogger(__name__)


class Task(object):
    def __init__(self, work_queue, priority, insert_number, promise):
        self.work_queue = work_queue
        self.priority = priority
        self.insert_number = insert_number
        self.promise = promise
        self.active = True
        self.created_by = traceback.extract_stack()

    def _on_error(self, exc, hide_error):
        if not hide_error:
            LOGGER.error("=== Promise was queued by ===")
            for (idx, stack_el) in enumerate(self.created_by):
                LOGGER.error(
                    "%2d: %20s: L%5d: %s",
                    idx, stack_el[0], stack_el[1], stack_el[2]
                )
        self.work_queue._run_next_promise_locked()
        raise exc

    def __lt__(self, o):
        if self.priority < o.priority:
            return True
        if self.priority > o.priority:
            return False

        if self.insert_number < o.insert_number:
            return True

        return False


class WorkQueue(object):
    def __init__(self, name, stop_on_quit, hide_uncatched):
        self.insert_number = 0

        self.name = name
        self.lock = threading.RLock()
        self.queue = []
        self.all_tasks = {}
        self.running = False
        self.stop_on_quit = stop_on_quit
        self.hide_uncatched = hide_uncatched

    def add_promise(self, promise, priority=0):
        self.insert_number += 1

        task = Task(self, -1 * priority, self.insert_number, promise)

        with self.lock:
            heapq.heappush(self.queue, task)
            assert (
                promise not in self.all_tasks or
                not self.all_tasks[promise].active
            )
            self.all_tasks[promise] = task

            if not self.running:
                self._run_next_promise()

    def _run_next_promise(self):
        self.running = True

        try:
            task = None
            while task is None or not task.active:
                task = heapq.heappop(self.queue)
                if task.active:
                    self.all_tasks.pop(task.promise)
        except IndexError:
            self.running = False
            return

        promise = task.promise.then(self._run_next_promise_locked)
        promise.catch(task._on_error, self.hide_uncatched)
        promise.schedule()

    def _run_next_promise_locked(self, *args, **kwargs):
        with self.lock:
            self._run_next_promise()

    def cancel(self, promise):
        try:
            with self.lock:
                task = self.all_tasks[promise]
                task.active = False
        except KeyError:
            LOGGER.debug(
                "Cannot cancel promise [%s]. (already running ?)", promise
            )

    def cancel_all(self):
        # reset the queue
        with self.lock:
            self.queue = []
            self.all_tasks = {}


class Plugin(PluginBase):
    def __init__(self):
        self.queues = {}

    def get_interfaces(self):
        return ['work_queue']

    def get_deps(self):
        return [
            {
               'interface': 'mainloop',
               'defaults': ['openpaperwork_core.mainloop.asyncio'],
            },
        ]

    def work_queue_create(
            self, queue_name, stop_on_quit=False, hide_uncatched=False):
        LOGGER.debug(
            "Creating work queue [%s] (stop_on_quit=%s)",
            queue_name, stop_on_quit
        )
        self.queues[queue_name] = WorkQueue(
            queue_name, stop_on_quit, hide_uncatched
        )
        return True

    def work_queue_add_promise(self, queue_name, promise, priority=0):
        if queue_name not in self.queues:
            return None
        self.queues[queue_name].add_promise(promise, priority)
        return True

    def work_queue_cancel(self, queue_name, promise):
        if queue_name not in self.queues:
            return None
        self.queues[queue_name].cancel(promise)
        return True

    def work_queue_cancel_all(self, queue_name):
        if queue_name not in self.queues:
            return None
        self.queues[queue_name].cancel_all()
        return True

    def mainloop_quit(self):
        # violent quit (does it ever happen ?)
        for queue in self.queues.values():
            queue.cancel_all()

    def mainloop_quit_graceful(self):
        for queue in self.queues.values():
            if queue.stop_on_quit:
                queue.cancel_all()

    def on_quit(self):
        self.mainloop_quit_graceful()



import logging
import os
import subprocess
import shutil

from typing import Optional

from .. import PluginBase


LOGGER = logging.getLogger(__name__)


class Plugin(PluginBase):
    def get_interfaces(self):
        return ['external_apps']

    def get_deps(self):
        return [
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_core.fs.python'],
            },
        ]

    def external_app_open_file(self, file_url):
        if shutil.which("xdg-open") is None:
            return None
        LOGGER.info("Opening %s with xdg-open", file_url)
        os.spawnlp(os.P_NOWAIT, 'xdg-open', 'xdg-open', file_url)
        return True

    def external_app_open_folder(self, folder_url):
        return self.external_app_open_file(folder_url)

    def external_app_can_send_as_attachment(self) -> bool:
        return shutil.which("xdg-email") is not None

    def external_app_send_as_attachment(self, file_url: str) -> Optional[bool]:
        if not self.external_app_can_send_as_attachment():
            return None
        LOGGER.info("Sending %s as attachment with xdg-email", file_url)
        assert file_url.startswith("file://")
        file_path = self.core.call_success("fs_unsafe", file_url)
        # xdg-email returns immediately, and we are interested in whether is
        # raises an exception
        try:
            subprocess.run(['xdg-email', '--attach', file_path], check=True)
        except subprocess.CalledProcessError as e:
            raise OSError(
                f"Failed to run xdg-email to send {file_path} as attachment"
            ) from e
        return True



import pycountry

import openpaperwork_core

from .. import _


LANGUAGES = {
    "English": _("English"),
    "French": _("French"),
    "German": _("German"),
}


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return ['i18n_lang']

    def i18n_lang_iso639_3_to_full(self, iso):
        attrs = [
            'iso639_3_code',
            'terminology',
            'alpha_3',
        ]
        for attr in attrs:
            try:
                r = pycountry.pycountry.languages.get(**{attr: iso})
                if r is None:
                    continue

                r = r.name

                if r in LANGUAGES:
                    return LANGUAGES[r]
                return r
            except (KeyError, UnicodeDecodeError):
                pass
        return iso



import logging
import re
import xml.etree.cElementTree as etree

import pyocr
import pyocr.builders

import openpaperwork_core

from . import util


LOGGER = logging.getLogger(__name__)

PAGE_FILENAME_FMT = "paper.{}.words"
PAGE_FILENAME_REGEX = re.compile(r"paper\.(\d+)\.words")


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 100

    def get_interfaces(self):
        return [
            "doc_text",
            "page_boxes",
            'pages',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_gtk.fs.gio'],
            },
        ]

    def doc_internal_get_mtime_by_url(self, out: list, doc_url):
        mtime = util.get_doc_mtime(self.core, doc_url, PAGE_FILENAME_REGEX)
        if mtime is None:
            return
        out.append(mtime)

    def page_internal_get_mtime_by_url(self, out: list, doc_url, page_idx):
        mtime = util.get_page_mtime(
            self.core, doc_url, page_idx, PAGE_FILENAME_FMT
        )
        if mtime is None:
            return
        out.append(mtime)

    def page_internal_get_hash_by_url(self, out: list, doc_url, page_idx):
        h = util.get_page_hash(self.core, doc_url, page_idx, PAGE_FILENAME_FMT)
        if h is None:
            return
        out.append(h)

    def doc_get_text_by_url(self, out: list, doc_url):
        doc_nb_pages = self.core.call_success(
            "doc_get_nb_pages_by_url", doc_url
        )
        if doc_nb_pages is None:
            return

        for page_idx in range(0, doc_nb_pages):
            text = self.page_get_text_by_url(doc_url, page_idx)
            if text is None:
                continue
            out.append(text)

    def page_reset_by_url(self, doc_url, page_idx):
        page_url = self.core.call_success(
            "fs_join", doc_url, PAGE_FILENAME_FMT.format(page_idx + 1)
        )
        if self.core.call_success("fs_exists", page_url) is None:
            return None
        self.core.call_all("fs_unlink", page_url, trash=False)

    def page_has_text_by_url(self, doc_url, page_idx):
        page_url = self.core.call_success(
            "fs_join", doc_url, PAGE_FILENAME_FMT.format(page_idx + 1)
        )
        if self.core.call_success("fs_exists", page_url) is None:
            return None
        text = self.page_get_text_by_url(doc_url, page_idx)
        if text is None:
            return None
        text = text.strip()
        # If the file exists and is valid, it takes precedence over text from,
        # for instance, a PDF file --> we return False instead of None
        return len(text) > 0

    def page_get_text_by_url(self, doc_url, page_idx):
        task = "hocr_load_page_text({} p{})".format(doc_url, page_idx)
        self.core.call_all("on_perfcheck_start", task)

        page_url = self.core.call_success(
            "fs_join", doc_url, PAGE_FILENAME_FMT.format(page_idx + 1)
        )

        file_desc = self.core.call_success("fs_open", page_url)
        if file_desc is None:
            self.core.call_all("on_perfcheck_stop", task)
            return None

        with file_desc:
            txt = file_desc.read().strip()
        try:
            tree = etree.XML(txt)
            for tag in tree.iter():
                tag_name = tag.tag.rsplit("}", 1)[-1]  # ignore namespace
                if tag_name != 'body':
                    continue
                txt = etree.tostring(tag, encoding='utf-8', method='text')
                break
            else:
                # No body ?!
                LOGGER.warning("No tag 'body' found in %s", page_url)
                txt = etree.tostring(tree, encoding='utf-8', method='text')
            if isinstance(txt, bytes):
                txt = txt.decode('utf-8')
            return txt
        except etree.ParseError as exc:
            LOGGER.warning(
                "%s contains invalid XML (%s). Will try with HTML parser",
                page_url, exc
            )

        def line_txt_generator(line):
            return " ".join(
                (word_box.content for word_box in line.word_boxes)
            )

        line_boxes = self.page_get_boxes_by_url(doc_url, page_idx)
        if line_boxes is None:
            return None
        return "\n".join(
            (line_txt_generator(line_box) for line_box in line_boxes)
        )

    def page_get_boxes_by_url(self, doc_url, page_idx):
        task = "hocr_load_page_boxes({} p{})".format(doc_url, page_idx)
        self.core.call_all("on_perfcheck_start", task)

        page_url = self.core.call_success(
            "fs_join", doc_url, PAGE_FILENAME_FMT.format(page_idx + 1)
        )

        file_desc = self.core.call_success("fs_open", page_url)
        if file_desc is None:
            self.core.call_all("on_perfcheck_stop", task)
            return None

        with file_desc:
            box_builder = pyocr.builders.LineBoxBuilder()
            boxes = box_builder.read_file(file_desc)
            if len(boxes) > 0:
                self.core.call_all("on_perfcheck_stop", task)
                return boxes

        with self.core.call_success("fs_open", page_url) as file_desc:
            # fallback: old format: word boxes
            # shouldn't be used anymore ...
            box_builder = pyocr.builders.WordBoxBuilder()
            boxes = box_builder.read_file(file_desc)
            if len(boxes) > 0:
                LOGGER.warning(
                    "Doc %s (page %d) uses old box format",
                    doc_url, page_idx
                )
                self.core.call_all("on_perfcheck_stop", task)
                return boxes

        self.core.call_all("on_perfcheck_stop", task)

    def page_set_boxes_by_url(self, doc_url, page_idx, boxes):
        page_url = self.core.call_success(
            "fs_join", doc_url, PAGE_FILENAME_FMT.format(page_idx + 1)
        )
        with self.core.call_success("fs_open", page_url, 'w') as file_desc:
            pyocr.builders.LineBoxBuilder().write_file(file_desc, boxes)
        return True

    def page_delete_by_url(self, doc_url, page_idx):
        return util.delete_page_file(
            self.core, PAGE_FILENAME_FMT, doc_url, page_idx
        )

    def page_move_by_url(
                self,
                source_doc_url, source_page_idx,
                dest_doc_url, dest_page_idx
            ):
        return util.move_page_file(
            self.core, PAGE_FILENAME_FMT,
            source_doc_url, source_page_idx,
            dest_doc_url, dest_page_idx
        )



#    Paperwork - Using OCR to grep dead trees the easy way
#    Copyright (C) 2012-2019  Jerome Flesch
#
#    Paperwork is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Paperwork is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Paperwork.  If not, see <http://www.gnu.org/licenses/>.
import sys

import openpaperwork_core


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        self.output = []

    def get_interfaces(self):
        return ['print']

    def print(self, txt):
        self.output.append(txt)

    def print_flush(self):
        output = "\n".join(self.output)
        sys.stdout.write(output)
        self.output = []



"""
Provides support for URIs "memory://". Those files are actually stored in
memory. It is only useful for temporary files. It has been made as a plugin
so it can easily be disabled on low-memory systems (will fall back on
gio.fs --> real on-disk files).
"""

import io
import itertools
import logging
import time

from . import CommonFsPluginBase


LOGGER = logging.getLogger(__name__)


class _MemoryFileAdapter(io.RawIOBase):
    def __init__(self, plugin, key, mode='r'):
        super().__init__()
        self.plugin = plugin
        self.mode = mode
        self.key = key

        self.io = None
        self.io_cls = io.BytesIO if 'b' in mode else io.StringIO

        if 'r' in mode or 'a' in mode:
            if key not in self.plugin.fs:
                raise FileNotFoundError(key)

            data = self.plugin.fs[key][1]

            if 'b' in mode and isinstance(data, str):
                data = data.encode("utf-8")
            elif 'b' not in mode and isinstance(data, bytes):
                data = data.decode("utf-8")

            self.io = self.io_cls(data)
        elif 'w' in mode:
            self.io = self.io_cls()
            self.plugin.fs[self.key] = (
                time.time(), b"" if 'b' in mode else ""
            )

        self.get_content = (
            self.io.getbuffer if 'b' in mode else self.io.getvalue
        )
        self.read = self.io.read if hasattr(self.io, 'read') else None
        self.readall = self.io.readall if hasattr(self.io, 'readall') else None
        self.readinto = (
            self.io.readinto if hasattr(self.io, 'readinto') else None
        )
        self.readline = (
            self.io.readline if hasattr(self.io, 'readline') else None
        )
        self.readlines = (
            self.io.readlines if hasattr(self.io, 'readlines') else None
        )
        self.seek = self.io.seek if hasattr(self.io, 'seek') else None
        self.tell = self.io.tell if hasattr(self.io, 'tell') else None
        self.truncate = (
            self.io.truncate if hasattr(self.io, 'truncate') else None
        )
        self.write = self.io.write if hasattr(self.io, 'write') else None
        self.writelines = (
            self.io.writelines if hasattr(self.io, 'writelines') else None
        )

    def readable(self):
        return 'r' in self.mode or '+' in self.mode

    def writable(self):
        return 'w' in self.mode or 'a' in self.mode

    def seekable(self):
        return True

    def fileno(self):
        raise io.UnsupportedOperation("fileno() called on memory object")

    def isatty(self):
        return False

    def writelines(self, lines):
        if 'b' in self.mode:
            self.write(b"".join(lines))
        else:
            self.write("".join(lines))

    def flush(self):
        super().flush()
        if hasattr(self.io, 'flush'):
            self.io.flush()
        if 'w' not in self.mode and 'a' not in self.mode:
            return
        if 'b' in self.mode:
            self.plugin.fs[self.key] = (time.time(), bytes(self.get_content()))
        else:
            self.plugin.fs[self.key] = (time.time(), str(self.get_content()))

    def close(self):
        self.flush()
        super().close()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()


class Plugin(CommonFsPluginBase):
    PRIORITY = 100  # to be called first for fs_mktemp

    def __init__(self):
        super().__init__()
        # self.fs = {id: (mtime, content), id: (mtime, content), ...}
        self.fs = {}
        self.id_gen = itertools.count()

    @staticmethod
    def _get_memory_id(uri):
        if not uri.startswith("memory://"):
            return None
        return uri[len("memory://"):]

    def fs_open(self, uri, mode='r', needs_fileno=False, **kwargs):
        if needs_fileno:
            return None
        mem_id = self._get_memory_id(uri)
        if mem_id is None:
            return None
        return _MemoryFileAdapter(self, mem_id, mode)

    def fs_exists(self, url):
        mem_id = self._get_memory_id(url)
        if mem_id is None:
            return None
        return mem_id in self.fs

    def fs_listdir(self, url):
        mem_id = self._get_memory_id(url)
        if mem_id is None:
            return

        mem_id += '/'
        out = []

        for k in self.fs.keys():
            if k.startswith(mem_id) and '/' not in k[len(mem_id):]:
                out.append(k)

        return out

    def fs_rename(self, old_url, new_url):
        old_mem_id = self._get_memory_id(old_url)
        new_mem_id = self._get_memory_id(old_url)

        if old_mem_id is None or new_mem_id is None:
            return

        f = self.fs.pop(old_mem_id)
        self.fs[new_mem_id] = f
        return True

    def fs_unlink(self, url, **kwargs):
        mem_id = self._get_memory_id(url)
        if mem_id is None:
            return
        self.fs.pop(mem_id)
        return True

    def fs_rm_rf(self, url, **kwargs):
        mem_id = self._get_memory_id(url)
        if mem_id is None:
            return

        if mem_id in self.fs:
            self.fs.pop(mem_id)

        mem_id += '/'
        for k in list(self.fs.keys()):
            if k.startswith(mem_id):
                self.fs.pop(k)

        return True

    def fs_get_mtime(self, url):
        mem_id = self._get_memory_id(url)
        if mem_id is None:
            return
        return self.fs[mem_id][0]

    def fs_getsize(self, url):
        mem_id = self._get_memory_id(url)
        if mem_id is None:
            return
        return len(self.fs[mem_id][1])

    def fs_isdir(self, url):
        mem_id = self._get_memory_id(url)
        if mem_id is None:
            return None

        mem_id += '/'
        for k in list(self.fs.keys()):
            if k.startswith(mem_id):
                return True
        return None

    def fs_copy(self, old_url, new_url):
        old_mem_id = self._get_memory_id(old_url)
        new_mem_id = self._get_memory_id(old_url)

        if old_mem_id is None and new_mem_id is None:
            return None

        if old_mem_id is None or new_mem_id is None:
            # One memory url, one local filesystem url
            # use the more generic and cross-FS method
            return super().fs_copy(old_url, new_url)

        self.fs[new_mem_id] = self.fs[old_mem_id]
        return new_url

    def fs_mkdir_p(self, url):
        mem_id = self._get_memory_id(url)
        if mem_id is None:
            return None
        # nothing to do actually
        return True

    def fs_recurse(self, parent_url, dir_included=False):
        mem_id = self._get_memory_id(parent_url)
        if mem_id is None:
            return

        mem_id += '/'
        out = []

        for k in self.fs.keys():
            if k.startswith(mem_id):
                out.append(k)

        return out

    def fs_hide(self, url):
        return None

    def fs_get_mime(self, url):
        return None

    def fs_iswritable(self, url):
        mem_id = self._get_memory_id(url)
        if mem_id is None:
            return None
        return True

    def fs_mktemp(
                self, prefix=None, suffix=None, mode='w+b', on_disk=False,
                **kwargs
            ):
        assert prefix is None or '/' not in prefix
        assert suffix is None or '/' not in suffix
        if on_disk:
            return
        name = "{}{}{}".format(
            prefix if prefix is not None else "",
            next(self.id_gen),
            suffix if suffix is not None else ""
        )
        return ("memory://" + name, _MemoryFileAdapter(self, name, mode))



import logging
import traceback

try:
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

import openpaperwork_core


from .. import deps


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.nb_dialogs = 3
        self.windows = []
        self.visible = False

    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_uncaught_exception',
            'gtk_window_listener',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'gtk_bug_report_dialog',
                'defaults': ['openpaperwork_gtk.bug_report'],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'uncaught_exception',
                'defaults': ['openpaperwork_core.uncaught_exception'],
            },
        ]

    def chkdeps(self, out: dict):
        if not GTK_AVAILABLE:
            out['gtk'].update(deps.GTK)

    def on_gtk_window_opened(self, window):
        self.windows.append(window)

    def on_gtk_window_closed(self, window):
        self.windows.remove(window)

    def on_uncaught_exception(self, exc_info):
        if exc_info[-1] is None:
            LOGGER.warning("No stacktrace. Won't display error dialog")
            return

        if self.visible:
            LOGGER.warning(
                "Error dialog currently displayed. Won't display another one"
            )
            return

        if self.nb_dialogs <= 0:
            LOGGER.warning(
                "Too many error dialogs displayed. Won't display them anymore"
            )
            return

        LOGGER.info("Uncaught exception. Showing error dialog")

        content = traceback.format_exception(*exc_info)
        content = ''.join(content)

        widget_tree = self.core.call_success(
            "gtk_load_widget_tree", "openpaperwork_gtk.uncaught_exception",
            "uncaught_exception.glade"
        )
        if widget_tree is None:
            LOGGER.warning("Failed to load widget tree")
            return

        widget_tree.get_object("error").set_text(content, -1)
        dialog = widget_tree.get_object("dialog")
        if len(self.windows) > 0:
            dialog.set_transient_for(self.windows[-1])
        dialog.connect("response", self._on_response)
        dialog.show_all()

        self.visible = True
        self.nb_dialogs -= 1

    def _on_response(self, dialog, response_id):
        LOGGER.info("Response: %s", response_id)

        dialog.destroy()
        self.visible = False

        if response_id != Gtk.ResponseType.ACCEPT:
            return

        self.core.call_all("open_bug_report")



import logging

import openpaperwork_core

from . import BaseDocViewController


LOGGER = logging.getLogger(__name__)


class BaseLayoutController(BaseDocViewController):
    def __init__(self, plugin):
        super().__init__(plugin)
        self.core = plugin.core
        self.real_nb_pages = self.core.call_success(
            "doc_get_nb_pages_by_url", self.plugin.active_doc[1]
        )
        if self.real_nb_pages is None:
            self.real_nb_pages = 0

    def on_close(self):
        for page in self.plugin.page_layout.get_children():

            # weird but has to be done. Otherwise it seems the children
            # are destroyed with 'page'
            for c in page.get_children():
                page.remove(c)

            self.plugin.page_layout.remove(page)
        self.plugin.pages = []
        self.plugin.widget_to_page = {}
        self.plugin.page_to_widget = {}

    def _update_visibility(self):
        vadj = self.plugin.scroll.get_vadjustment()
        lower = vadj.get_lower()
        p_min = vadj.get_value() - lower
        p_max = vadj.get_value() + vadj.get_page_size() - lower
        for widget in self.plugin.page_layout.get_children():
            alloc = widget.get_allocation()
            p_lower = alloc.y
            p_upper = alloc.y + alloc.height
            visible = (p_min <= p_upper and p_lower <= p_max)
            page = self.plugin.widget_to_page[widget]
            page.set_visible(visible)

    def on_vscroll_value_changed(self, vadj):
        super().on_vscroll_value_changed(vadj)
        self._update_visibility()

    def on_vscroll_changed(self, vadj):
        super().on_vscroll_changed(vadj)
        self._update_visibility()

    def on_page_size_obtained(self, page):
        super().on_page_size_obtained(page)
        self._update_visibility()

    def on_layout_size_allocate(self, layout):
        super().on_layout_size_allocate(layout)
        self._update_visibility()

    def _add_pages(self):
        pages = []
        self.core.call_all(
            "doc_open_components", pages, *self.plugin.active_doc
        )
        for (visible, page) in pages:
            self.plugin.docview_add_page_viewer(page, visible)

    def doc_reload(self):
        super().doc_reload()
        LOGGER.info("Reloading document %s", self.plugin.active_doc)
        self.on_close()
        self._add_pages()
        for page in self.plugin.pages:
            page.load()


class LayoutControllerLoading(BaseLayoutController):
    def __init__(self, plugin):
        super().__init__(plugin)
        self.nb_loaded = 0

        # page instantiation must be done before the calls to enter()
        # because other controllers may need the pages. So we cheat a little
        # bit and make it in the constructor.
        self._add_pages()

    def enter(self):
        super().enter()
        if len(self.plugin.pages) <= 0:
            self.plugin.docview_switch_controller(
                'layout', LayoutControllerLoaded
            )
            return
        self.nb_loaded = 0
        for page in self.plugin.pages:
            page.load()

    def on_page_size_obtained(self, page):
        super().on_page_size_obtained(page)
        self.nb_loaded += 1
        if self.nb_loaded >= len(self.plugin.pages):
            self.plugin.docview_switch_controller(
                'layout', LayoutControllerLoaded
            )

    def exit(self):
        LOGGER.info(
            "Size of all pages of doc %s loaded", self.plugin.active_doc[0]
        )


class LayoutControllerLoaded(BaseLayoutController):
    def enter(self):
        super().enter()
        self._update_visibility()


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return ['gtk_docview_controller']

    def get_deps(self):
        return [
            {
                'interface': 'gtk_docview',
                'defaults': ['paperwork_gtk.mainwindow.docview'],
            },
        ]

    def gtk_docview_get_controllers(self, out: dict, docview):
        out['layout'] = LayoutControllerLoading(docview)



import json
import logging

import openpaperwork_core


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return ['authors']

    def get_deps(self):
        return [
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_gtk.fs.gio'],
            },
            {
                'interface': 'resources',
                'defaults': ['openpaperwork_core.resources.setuptools'],
            },
        ]

    def authors_get(self, out: dict):
        file_path_ctx = self.core.call_success(
            "resources_get_file", "paperwork_backend.authors",
            "AUTHORS.json"
        )
        if file_path_ctx is None:
            LOGGER.error("AUTHORS.json is missing !")
            return None
        with file_path_ctx as file_path:
            file_path = self.core.call_success("fs_safe", file_path)
            with self.core.call_success("fs_open", file_path, 'r') as fd:
                content = fd.read()
        content = json.loads(content)

        for category in content:
            for (category_name, authors) in category.items():
                out[category_name] = authors

        return True



import datetime
import itertools
import logging
import math

import openpaperwork_core
import openpaperwork_core.deps


LOGGER = logging.getLogger(__name__)

PDF_FILENAME = 'doc.pdf'
PASSWD_FILENAME = 'passwd.txt'
MAPPING_FILE = "page_map.csv"

PDF_RENDER_FACTOR = 4


def minmax_rects(rects):
    (mx1, my1, mx2, my2) = (math.inf, math.inf, 0, 0)
    for rectangle in rects:
        ((x1, y1), (x2, y2)) = (
            (int(rectangle.x1 * PDF_RENDER_FACTOR),
             int(rectangle.y2 * PDF_RENDER_FACTOR)),
            (int(rectangle.x2 * PDF_RENDER_FACTOR),
             int(rectangle.y1 * PDF_RENDER_FACTOR))
        )
        (x1, x2) = (min(x1, x2), max(x1, x2))
        (y1, y2) = (min(y1, y2), max(y1, y2))
        mx1 = min(mx1, x1)
        my1 = min(my1, y1)
        mx2 = max(mx2, x2)
        my2 = max(my2, y2)
    rect = ((mx1, my1), (mx2, my2))
    return rect


class PdfWordBox(object):
    def __init__(self, content, position):
        self.content = content
        self.position = minmax_rects(position)

    def __str__(self):
        return "{{ .position={}, .content={} }}".format(
            self.position, str(self.content)
        )

    def __lt__(self, o):
        return self.position < o.position

    def __repr__(self):
        return str(self)


class PdfLineBox(object):
    def __init__(self, word_boxes, position):
        self.word_boxes = word_boxes
        self.position = minmax_rects(position)

    def __str__(self):
        return "{{ .position={}, .word_boxes={} }}".format(
            self.position, str(self.word_boxes)
        )

    def __repr__(self):
        return str(self)

    def _get_content(self):
        return " ".join([w.content for w in self.word_boxes])

    content = property(_get_content)


class PdfPageMapping(object):
    SEPARATOR = ","

    def __init__(self, plugin, doc_url):
        self.plugin = plugin
        self.core = plugin.core
        self.doc_url = doc_url

        self.map_url = self.core.call_success(
            "fs_join", self.doc_url, MAPPING_FILE
        )

        self.mapping = None
        self.reverse_mapping = None
        self.page_mtimes = None

        self.nb_pages = -1
        self.real_nb_pages = -1

    def set_mapping(self, original_page_idx, target_page_idx):
        """
        Indicates that the page 'original_page_idx' in the actual PDF file
        is displayed as being the page 'target_page_idx'.
        If 'target_page_idx' < 0, it means the page is not displayed anymore
        (--> act if it is deleted)
        """
        if original_page_idx >= self.real_nb_pages:
            # comes from another set of plugins, no point in tracking it
            return

        now = datetime.datetime.now().timestamp()

        if target_page_idx is not None:
            old_original = self.reverse_mapping.get(target_page_idx, None)
            if old_original is not None and old_original in self.mapping:
                self.mapping.pop(old_original)

        old_target = self.mapping.get(original_page_idx, None)
        if old_target is not None and old_target in self.reverse_mapping:
            self.reverse_mapping.pop(old_target)
            self.page_mtimes[old_target] = now

        if target_page_idx is None and original_page_idx in self.mapping:
            self.mapping.pop(original_page_idx)
        else:
            self.mapping[original_page_idx] = target_page_idx

        if target_page_idx is not None:
            self.reverse_mapping[target_page_idx] = original_page_idx
            self.page_mtimes[target_page_idx] = now

    def update_nb_pages(self):
        self.nb_pages = 0
        if len(self.mapping) <= 0:
            return
        for v in self.mapping.values():
            if v is None:
                continue
            if v >= self.nb_pages:
                self.nb_pages = v + 1

    def load(self):
        try:
            self.real_nb_pages = self.plugin._doc_internal_get_nb_pages_by_url(
                self.doc_url, mapping=False
            )

            self.mapping = {p: p for p in range(0, self.real_nb_pages)}
            self.reverse_mapping = {p: p for p in range(0, self.real_nb_pages)}
            now = datetime.datetime.now().timestamp()
            self.page_mtimes = {p: now for p in range(0, self.real_nb_pages)}
            if self.core.call_success("fs_exists", self.map_url) is None:
                self.update_nb_pages()
                return

            with self.core.call_success("fs_open", self.map_url, "r") as fd:
                # drop the first line
                lines = fd.readlines()[1:]
                lines = [line.split(",", 1) for line in lines]
                lines = [
                    (int(orig_page_idx), int(target_page_idx))
                    for (orig_page_idx, target_page_idx) in lines
                ]
                lines.sort()
                for (orig_page_idx, target_page_idx) in lines:
                    if target_page_idx < 0:
                        target_page_idx = None
                    self.set_mapping(orig_page_idx, target_page_idx)

            self.update_nb_pages()
        except Exception as exc:
            LOGGER.critical(
                "Failed to load page mapping for doc %s !",
                self.doc_url, exc_info=exc
            )
            raise

    def load_reverse_only(self):
        """
        Load the mapping, but doesn't look at the total number of pages
        in the document. Avoid opening the PDF file.
        """
        self.mapping = None
        self.reverse_mapping = {}

        if self.core.call_success("fs_exists", self.map_url) is None:
            return

        with self.core.call_success("fs_open", self.map_url, "r") as fd:
            # drop the first line
            lines = fd.readlines()[1:]
            for line in lines:
                (orig_page_idx, target_page_idx) = line.split(",", 1)
                orig_page_idx = int(orig_page_idx)
                target_page_idx = int(target_page_idx)
                if target_page_idx < 0:
                    continue
                self.reverse_mapping[target_page_idx] = orig_page_idx

    def save(self):
        if self.core.call_success("fs_isdir", self.doc_url) is None:
            return

        nb_maps = 0
        for (orig_page_idx, target_page_idx) in self.mapping.items():
            if orig_page_idx == target_page_idx:
                continue
            nb_maps += 1

        if nb_maps <= 0:
            if self.core.call_success("fs_exists", self.map_url) is not None:
                self.core.call_success("fs_unlink", self.map_url)
            return

        with self.core.call_success("fs_open", self.map_url, "w") as fd:
            fd.write("original_page_index,target_page_index\n")
            mapping = list(self.mapping.items())
            mapping.sort()
            for (orig_page_idx, target_page_idx) in mapping:
                if orig_page_idx == target_page_idx:
                    continue
                fd.write("{},{}\n".format(orig_page_idx, target_page_idx))

            for page_idx in range(0, self.real_nb_pages):
                if page_idx not in self.mapping:
                    fd.write("{},-1\n".format(page_idx))

    def has_original_page_idx(self, original_page_idx):
        if self.mapping is None:
            self.load()
        return original_page_idx in self.mapping

    def get_original_page_idx(self, target_page_idx, reverse_only=False):
        if reverse_only:
            if self.reverse_mapping is None:
                self.load_reverse_only()
        else:
            if self.mapping is None:
                self.load()
        # Keep in mind we may have loaded only the mapping --> we don't know
        # how many pages there are in the PDF, so the mapping may be incomplete
        original_page_idx = self.reverse_mapping.get(target_page_idx, None)
        return original_page_idx

    def get_target_page_mtime(self, target_page_idx):
        if self.mapping is None:
            self.load()
        return self.page_mtimes.get(target_page_idx, None)

    def get_target_page_hash(self, target_page_idx):
        if self.mapping is None:
            self.load()
        original_page_idx = self.reverse_mapping.get(
            target_page_idx, None
        )
        if original_page_idx is None:
            return None
        return hash(original_page_idx)

    def _move_pages(self, original_page_idx, target_page_idx, offset):
        mapping = list(self.mapping.items())
        mapping.sort(key=lambda x: x[1], reverse=offset > 0)
        for (original, target) in mapping:
            if target >= target_page_idx:
                LOGGER.info(
                    "Page %d (original) ; target: %d --> %d",
                    original, target, target + offset
                )
                self.set_mapping(original, target + offset)

        self.update_nb_pages()

    def delete_target_page(self, target_page_idx):
        if self.mapping is None:
            self.load()

        LOGGER.info(
            "Deleting page %d from PDF %s",
            target_page_idx, self.doc_url
        )
        original_page_idx = self.reverse_mapping.pop(
            target_page_idx, None
        )
        if original_page_idx is not None:
            self.mapping.pop(original_page_idx, None)
        else:
            original_page_idx = target_page_idx

        self._move_pages(original_page_idx, target_page_idx, offset=-1)

    def make_room_for_target_page(self, target_page_idx):
        if self.mapping is None:
            self.load()
        original_page_idx = self.reverse_mapping.get(
            target_page_idx, target_page_idx
        )
        self._move_pages(original_page_idx, target_page_idx, offset=1)

    def print_mapping(self):
        if self.mapping is None:
            self.load()

        nb_pages = self.plugin._doc_internal_get_nb_pages_by_url(
            self.doc_url, mapping=False
        )

        print("==== MAPPING OF {} (nb_pages={}|{}) ====".format(
            self.doc_url, nb_pages, self.nb_pages
        ))
        for original_page_idx in range(0, nb_pages):
            target_page_idx = self.mapping.get(original_page_idx, None)
            if target_page_idx is None:
                print("{} --> {}".format(original_page_idx, original_page_idx))
                continue
            print("{} --> {}".format(original_page_idx, target_page_idx))
            if target_page_idx < 0:
                continue
            if target_page_idx not in self.reverse_mapping:
                print("WARNING: NO REVERSE")
            else:
                if original_page_idx != self.reverse_mapping[target_page_idx]:
                    print("WARNING: REVERSE DOESN'T MATCH: {} <-- {}".format(
                        target_page_idx, original_page_idx
                    ))
        print("=======================")

    def get_map_hash(self):
        if self.core.call_success("fs_exists", self.map_url) is None:
            return None
        return self.core.call_success("fs_hash", self.map_url)

    def get_map_mtime(self):
        if self.core.call_success("fs_exists", self.map_url) is None:
            return None
        return self.core.call_success("fs_get_mtime", self.map_url)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 20

    def __init__(self):
        super().__init__()
        # we cache the hash of PDF files since they never change
        self.cache_hash = {}
        # we cache the number of pages in PDF files since they never change
        # (real number before mapping)
        self.cache_nb_pages = {}

        self.cache_mappings = {}
        self.cache_passwords = {}

    def get_interfaces(self):
        return [
            "doc_hash",
            "doc_pdf",
            "doc_pdf_import",
            "doc_pdf_url",
            "doc_text",
            "doc_type",
            "page_boxes",
            "page_img",
            "page_paper",
            'pages',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_gtk.fs.gio'],
            },
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_gtk.mainloop.glib'],
            },
            {
                # to provide doc_get_nb_pages_by_url()
                'interface': 'nb_pages',
                'defaults': ['paperwork_backend.model'],
            },
            # for page_move_by_url():
            {
                'interface': 'pillow',
                'defaults': [
                    'openpaperwork_core.pillow.img',
                    'paperwork_backend.pillow.pdf',
                ],
            },
            {
                'interface': 'poppler',
                'defaults': ['paperwork_backend.poppler.memory'],
            },
            {
                'interface': 'urls',
                'defaults': ['openpaperwork_core.urls'],
            },
        ]

    def _get_pdf_url(self, doc_url):
        if doc_url.endswith(".pdf"):
            return doc_url
        pdf_url = self.core.call_success("fs_join", doc_url, PDF_FILENAME)
        if self.core.call_success("fs_exists", pdf_url) is None:
            return None
        return pdf_url

    def _get_pdf_password(self, doc_url):
        password = self.cache_passwords.get(doc_url, None)
        if password is not None:
            return password

        passwd_url = self.core.call_success(
            "fs_join", doc_url, PASSWD_FILENAME
        )
        if self.core.call_success("fs_exists", passwd_url):
            with self.core.call_success("fs_open", passwd_url, "r") as fd:
                password = fd.read().strip()
        self.cache_passwords[doc_url] = password
        return password

    def _get_page_mapping(self, doc_url):
        if doc_url in self.cache_mappings:
            return self.cache_mappings[doc_url]
        mapping = PdfPageMapping(self, doc_url)
        self.cache_mappings[doc_url] = mapping
        return mapping

    def _open_pdf(self, doc_url):
        pdf_url = self._get_pdf_url(doc_url)
        if pdf_url is None:
            return (None, None)

        password = self._get_pdf_password(doc_url)
        LOGGER.info("Opening %s (password=%s)", pdf_url, bool(password))
        doc = self.core.call_success(
            "poppler_open", pdf_url, password=password
        )
        return (pdf_url, doc)

    def is_doc(self, doc_url):
        pdf_url = self._get_pdf_url(doc_url)
        if pdf_url is None:
            return None
        return True

    def doc_get_pdf_url_by_url(self, doc_url, write=False, allow_mapped=False):
        if write:
            return self.core.call_success("fs_join", doc_url, PDF_FILENAME)
        if not allow_mapped:
            mapping_url = self.core.call_success(
                "fs_join", doc_url, MAPPING_FILE
            )
            if self.core.call_success("fs_exists", mapping_url):
                return None
        return self._get_pdf_url(doc_url)

    def flush_doc_cache(self, doc_url):
        self.cache_hash.pop(doc_url, None)
        self.cache_passwords.pop(doc_url, None)
        self.cache_mappings.pop(doc_url, None)
        self.cache_nb_pages.pop(doc_url, None)

    def doc_get_hash_by_url(self, doc_url):
        pdf_url = self._get_pdf_url(doc_url)
        if pdf_url is None:
            return

        # cache the hash of doc.pdf to speed up imports
        if doc_url not in self.cache_hash:
            h = self.core.call_success("fs_hash", pdf_url)
            self.cache_hash[doc_url] = h
        return self.cache_hash[doc_url]

    def page_internal_get_hash_by_url(self, out: list, doc_url, page_idx):
        pdf_url = self._get_pdf_url(doc_url)
        if pdf_url is None:
            return
        mapping = self._get_page_mapping(doc_url)
        page_hash = mapping.get_target_page_hash(page_idx)
        if page_hash is None:
            # deleted page or handled by another plugin
            return
        out.append(page_hash)

        # cache the hash of doc.pdf to speed up imports
        if doc_url not in self.cache_hash:
            h = self.core.call_success("fs_hash", pdf_url)
            self.cache_hash[doc_url] = h
        out.append(self.cache_hash[doc_url])

    def doc_internal_get_mtime_by_url(self, out: list, doc_url):
        pdf_url = self._get_pdf_url(doc_url)
        if pdf_url is None:
            return None
        mapping = self._get_page_mapping(doc_url)
        mtime = mapping.get_map_mtime()
        if mtime is not None:
            out.append(mtime)

        mtime = self.core.call_success("fs_get_mtime", pdf_url)
        if mtime is None:
            return None
        out.append(mtime)

    def page_internal_get_mtime_by_url(self, out: list, doc_url, page_idx):
        pdf_url = self._get_pdf_url(doc_url)
        if pdf_url is None:
            return

        mapping = self._get_page_mapping(doc_url)
        mtime = mapping.get_target_page_mtime(page_idx)
        if mtime is not None:
            out.append(mtime)

        mtime = self.core.call_success("fs_get_mtime", pdf_url)
        if mtime is not None:
            out.append(mtime)
        return

    def _doc_get_real_nb_pages_by_url(self, doc_url):
        (pdf_url, pdf) = self._open_pdf(doc_url)
        if pdf is None:
            return None
        nb_pages = pdf.get_n_pages()
        return nb_pages

    def doc_pdf_get_real_nb_pages_by_url(self, doc_url):
        # Poppler is not thread-safe
        return self.core.call_one(
            "mainloop_execute", self._doc_get_real_nb_pages_by_url, doc_url
        )

    def _doc_internal_get_nb_pages_by_url(
            self, doc_url, mapping=True):
        if mapping:
            pdf_url = self._get_pdf_url(doc_url)
            if pdf_url is None:
                return 0
            mapping = self._get_page_mapping(doc_url)
            if mapping.nb_pages < 0:
                mapping.load()
            return mapping.nb_pages

        if doc_url in self.cache_nb_pages:
            r = self.cache_nb_pages[doc_url]
        else:
            r = self.doc_pdf_get_real_nb_pages_by_url(doc_url)
            if r is not None:
                self.cache_nb_pages[doc_url] = r
        return r if r is not None else 0

    def doc_internal_get_nb_pages_by_url(self, out: list, doc_url):
        r = self._doc_internal_get_nb_pages_by_url(doc_url)
        if r == 0:
            return
        out.append(r)

    def page_get_img_url(self, doc_url, page_idx, write=False, **kwargs):
        if write:
            return None
        pdf_url = self._get_pdf_url(doc_url)
        if pdf_url is None:
            return None

        # The first page is requested much more often than the others, due
        # to thumbnailing.
        # ASSUMPTION(Jflesch): If there is a doc.pdf, there is a first page
        # in it.
        # We don't want to open the PDF file for each thumbnail (first page)
        # but at the same time, we still need to check the mapping file.

        mapping = self._get_page_mapping(doc_url)
        original_page_idx = mapping.get_original_page_idx(
            page_idx, reverse_only=(page_idx == 0)
        )
        if original_page_idx is None:
            if page_idx != 0 and page_idx >= mapping.nb_pages:
                return None
            original_page_idx = page_idx

        password = self._get_pdf_password(doc_url)
        if password is not None:
            password = password.encode("utf-8").hex()

        # same URL used in browsers
        url = self.core.call_success(
            "url_args_join", pdf_url,
            page=str(original_page_idx + 1),
            password=password
        )
        return url

    @staticmethod
    def _custom_split(input_str, input_rects, splitter, log_txt):
        # turn text and layout from Poppler into boxes
        # XXX(Jflesch): following assert fails sometimes ? oO
        # assert len(input_str) == len(input_rects)
        if len(input_str) != len(input_rects):
            LOGGER.warning(
                "%s: Input strings: %d ; Input rects: %d",
                log_txt, len(input_str), len(input_rects)
            )
            m = min(len(input_str), len(input_rects))
            input_str = input_str[:m]
            input_rects = input_rects[:m]
        input_el = zip(input_str, input_rects)
        for (is_split, group) in itertools.groupby(
                    input_el,
                    lambda x: splitter(x[0])
                ):
            if is_split:
                continue
            letters = ""
            rects = []
            for (letter, rect) in group:
                letters += letter
                rects.append(rect)
            yield (letters, rects)

    def _doc_get_text_by_url(self, out: list, doc_url):
        task = "pdf_get_text_by_url({})".format(doc_url)
        self.core.call_all("on_perfcheck_start", task)

        (pdf_url, pdf) = self._open_pdf(doc_url)
        if pdf is None:
            self.core.call_all("on_perfcheck_stop", task)
            return
        mapping = self._get_page_mapping(doc_url)

        for page_idx in range(0, pdf.get_n_pages()):
            if not mapping.has_original_page_idx(page_idx):
                continue
            page = pdf.get_page(page_idx)
            # some PDF are really badly damaged
            if page is None:
                continue
            try:
                txt = page.get_text()
                txt = txt.strip()
            except UnicodeDecodeError as exc:
                LOGGER.warning(
                    "%s p%d: UnicodeDecodeError: Assuming page has no text",
                    doc_url, page_idx, exc_info=exc
                )
                continue
            if txt == "":
                continue
            out.append(txt)
        self.core.call_all(
            "on_perfcheck_stop", task, nb_pages=pdf.get_n_pages()
        )
        return True

    def doc_get_text_by_url(self, out: list, doc_url):
        # Poppler is not thread-safe
        return self.core.call_one(
            "mainloop_execute",
            self._doc_get_text_by_url, out, doc_url
        )

    def _page_has_text_by_url(self, doc_url, page_idx):
        if doc_url in self.cache_nb_pages:
            if page_idx >= self.cache_nb_pages[doc_url]:
                return None

        (pdf_url, pdf) = self._open_pdf(doc_url)
        if pdf is None:
            return None

        page = pdf.get_page(page_idx)
        if page is None:
            return None

        try:
            return len(page.get_text().strip()) > 0
        except UnicodeDecodeError as exc:
            LOGGER.warning(
                "%s p%d: UnicodeDecodeError: Assuming page has no text",
                doc_url, page_idx, exc_info=exc
            )
            return False

    def page_has_text_by_url(self, doc_url, page_idx):
        pdf_url = self._get_pdf_url(doc_url)
        if pdf_url is None:
            return None
        mapping = self._get_page_mapping(doc_url)
        page_idx = mapping.get_original_page_idx(page_idx)
        if page_idx is None:
            return None
        # Poppler is not thread-safe
        return self.core.call_one(
            "mainloop_execute", self._page_has_text_by_url, doc_url, page_idx
        )

    def _page_get_text_by_url(self, doc_url, page_idx):
        if doc_url in self.cache_nb_pages:
            if page_idx >= self.cache_nb_pages[doc_url]:
                return None

        (pdf_url, pdf) = self._open_pdf(doc_url)
        if pdf is None:
            return None

        page = pdf.get_page(page_idx)
        if page is None:
            return None

        try:
            return page.get_text().strip()
        except UnicodeDecodeError as exc:
            LOGGER.warning(
                "%s p%d: UnicodeDecodeError: Assuming page has no text",
                doc_url, page_idx, exc_info=exc
            )
            return False

    def page_get_text_by_url(self, doc_url, page_idx):
        pdf_url = self._get_pdf_url(doc_url)
        if pdf_url is None:
            return None
        mapping = self._get_page_mapping(doc_url)
        page_idx = mapping.get_original_page_idx(page_idx)
        if page_idx is None:
            return None
        # Poppler is not thread-safe
        return self.core.call_one(
            "mainloop_execute", self._page_get_text_by_url, doc_url, page_idx
        )

    def _page_get_boxes_by_url(self, doc_url, page_idx):
        if doc_url in self.cache_nb_pages:
            if page_idx >= self.cache_nb_pages[doc_url]:
                return None

        (pdf_url, pdf) = self._open_pdf(doc_url)
        if pdf is None:
            return

        pdf_page = pdf.get_page(page_idx)
        if pdf_page is None:
            return

        try:
            txt = pdf_page.get_text()
            if txt.strip() == "":
                return None
        except UnicodeDecodeError as exc:
            LOGGER.warning(
                "%s p%d: UnicodeDecodeError: Assuming page has no text",
                doc_url, page_idx, exc_info=exc
            )
            return None

        layout = pdf_page.get_text_layout()
        if not layout[0]:
            return None
        layout = layout[1]

        line_boxes = []
        for (line, line_rects) in self._custom_split(
                    txt, layout, lambda x: x == "\n", "lines",
                ):
            words = []
            for (word, word_rects) in self._custom_split(
                        line, line_rects, lambda x: x.isspace(), "words"
                    ):
                word_box = PdfWordBox(word, word_rects)
                words.append(word_box)
            line_boxes.append(PdfLineBox(words, line_rects))
        return line_boxes

    def page_get_boxes_by_url(self, doc_url, page_idx):
        pdf_url = self._get_pdf_url(doc_url)
        if pdf_url is None:
            return None
        mapping = self._get_page_mapping(doc_url)
        page_idx = mapping.get_original_page_idx(page_idx)
        if page_idx is None:
            return None
        # Poppler is not thread-safe
        return self.core.call_one(
            "mainloop_execute", self._page_get_boxes_by_url, doc_url, page_idx
        )

    def doc_pdf_import(self, src_file_uri, password=None, target_doc_id=None):
        doc_id = None
        doc_url = None

        if target_doc_id is not None:
            doc_id = target_doc_id
            doc_url = self.core.call_success(
                "doc_id_to_url", doc_id, existing=False
            )
            if doc_url is None:
                doc_id = None
            elif self.core.call_success("fs_exists", doc_url):
                doc_id = None

        if doc_id is None:
            (doc_id, doc_url) = self.core.call_success("storage_get_new_doc")

        # just to be safe
        self.cache_mappings.pop(doc_url, None)

        pdf_url = self.core.call_success("fs_join", doc_url, PDF_FILENAME)

        self.core.call_success("fs_mkdir_p", doc_url)
        self.core.call_success("fs_copy", src_file_uri, pdf_url)

        if password is not None:
            passwd_url = self.core.call_success(
                "fs_join", doc_url, PASSWD_FILENAME
            )
            with self.core.call_success("fs_open", passwd_url, "w") as fd:
                fd.write(password)

        try:
            # check the PDF is readable
            doc = self.core.call_success(
                "poppler_open", pdf_url, password=password
            )
            exc_info = None
        except Exception as exc:
            doc = None
            exc_info = exc

        if doc is None:
            if exc_info is None:
                LOGGER.error("Failed to read %s", pdf_url)
            else:
                LOGGER.error("Failed to read %s", pdf_url, exc_info=exc_info)
            self.core.call_success("fs_rm_rf", doc_url, trash=False)
            return (None, None)

        return (doc_id, doc_url)

    def page_delete_by_url(self, doc_url, page_idx):
        pdf_url = self._get_pdf_url(doc_url)
        if pdf_url is None:
            return
        mapping = self._get_page_mapping(doc_url)
        mapping.delete_target_page(page_idx)
        mapping.save()

    def _page_get_paper_size_by_url(self, doc_url, page_idx):
        if doc_url in self.cache_nb_pages:
            if page_idx >= self.cache_nb_pages[doc_url]:
                return None

        (pdf_url, pdf) = self._open_pdf(doc_url)
        if pdf is None:
            return None

        page = pdf.get_page(page_idx)
        if page is None:
            return

        size = page.get_size()

        # points --> inches: / 72
        # inches --> millimeters (i18n unit): * 25.4
        return (
            size[0] / 72.0 * 25.4,
            size[0] / 72.0 * 25.4,
        )

    def page_get_paper_size_by_url(self, doc_url, page_idx):
        pdf_url = self._get_pdf_url(doc_url)
        if pdf_url is None:
            return None
        mapping = self._get_page_mapping(doc_url)
        page_idx = mapping.get_original_page_idx(page_idx)
        if page_idx is None:
            return None
        # Poppler is not thread-safe
        return self.core.call_one(
            "mainloop_execute",
            self._page_get_paper_size_by_url, doc_url, page_idx
        )

    def page_move_by_url(
                self,
                source_doc_url, target_source_page_idx,
                dest_doc_url, target_dest_page_idx
            ):
        # 'source' is about the source document.
        # 'dest' is about the destination document.
        # 'original' is about the original position of a page in a PDF file.
        # 'target' is about the position of a page as shown to the end-user.

        source_pdf_url = self._get_pdf_url(source_doc_url)
        source_is_pdf = source_pdf_url is not None
        dest_is_pdf = self._get_pdf_url(dest_doc_url) is not None
        if not source_is_pdf and not dest_is_pdf:
            return

        if source_is_pdf:
            source_mapping = self._get_page_mapping(source_doc_url)
        if dest_is_pdf:
            dest_mapping = self._get_page_mapping(dest_doc_url)

        LOGGER.info(
            "%s (%s) p%d --> %s (%s) p%d",
            source_doc_url, "PDF" if source_is_pdf else "non-PDF",
            target_source_page_idx,
            dest_doc_url, "PDF" if dest_is_pdf else "non-PDF",
            target_dest_page_idx,
        )

        if source_is_pdf:
            original_source_page_idx = source_mapping.get_original_page_idx(
                target_source_page_idx
            )
            # if original_source_page_idx is None means
            # this page is not handled by us ; still, we must shift
            # all our pages
            LOGGER.info(
                "- Removing page (original=%s, target=%d) from %s",
                original_source_page_idx, target_source_page_idx,
                source_doc_url
            )
            source_mapping.delete_target_page(target_source_page_idx)

        if dest_is_pdf:
            LOGGER.info(
                "- Making room for a new page (target=%d) in %s",
                target_dest_page_idx, dest_doc_url
            )
            dest_mapping.make_room_for_target_page(target_dest_page_idx)

        if source_doc_url == dest_doc_url:
            assert source_mapping is dest_mapping
            if original_source_page_idx is not None:
                LOGGER.info(
                    "New mapping: %s: original=p%d --> target=p%d",
                    source_doc_url, original_source_page_idx,
                    target_dest_page_idx
                )
                source_mapping.set_mapping(
                    original_source_page_idx, target_dest_page_idx
                )

                source_mapping.update_nb_pages()
        elif source_is_pdf:
            if original_source_page_idx is not None:
                # export the PDF page as an image file
                # it relies on other model plugins (interface 'page_img'), but
                # they can't be declared as dependencies, as we do provide
                # 'page_img' too. It would make a dependency loop.

                # we are a low priority plugin: other plugins should
                # already have made room for our page if required

                source_img = "{}#page={}".format(
                    source_pdf_url, str(original_source_page_idx + 1)
                )
                LOGGER.info("Generating image from %s", source_img)
                source_img = self.core.call_success(
                    "url_to_pillow", source_img
                )

                # since PDF are not writable by themselves, we can call
                # page_get_img_url(write=True). We are sure that our
                # implementation of this method won't reply
                dest_img = self.core.call_success(
                    "page_get_img_url", dest_doc_url, target_dest_page_idx,
                    write=True, original=True
                )
                LOGGER.info("Writting page image back as %s", dest_img)
                self.core.call_success("pillow_to_url", source_img, dest_img)

        if source_is_pdf:
            source_mapping.save()
        if dest_is_pdf:
            dest_mapping.save()



#    Paperwork - Using OCR to grep dead trees the easy way
#    Copyright (C) 2012-2019  Jerome Flesch
#
#    Paperwork is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Paperwork is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Paperwork.  If not, see <http://www.gnu.org/licenses/>.
import logging

import openpaperwork_core

import rich.progress


LOGGER = logging.getLogger(__name__)

TIME_BETWEEN_PROGRESS = 0.3


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        self.console = None
        self.progress = None
        self.tasks = {}

    def get_interfaces(self):
        return [
            'progress_listener',
        ]

    def cmd_set_console(self, console):
        if console is None:
            return
        self.console = console

    def on_progress(self, upd_type, progress, description=None):
        if self.progress is None:
            self.progress = rich.progress.Progress(
                console=self.console.console
            )
            self.progress.start()
            return
        if upd_type not in self.tasks:
            if progress >= 1.0:
                return
            assert description is not None
            self.tasks[upd_type] = self.progress.add_task(
                description=description, total=1.0
            )
        if progress >= 1.0:
            task_id = self.tasks.pop(upd_type)
            self.progress.remove_task(task_id)
        else:
            task_id = self.tasks[upd_type]
            self.progress.update(task_id, completed=progress)

    def on_quit(self):
        if self.progress is None:
            return
        self.progress.stop()



import collections
import logging

try:
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

import openpaperwork_core
import openpaperwork_gtk.deps


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.windows = []
        self.groups = collections.defaultdict(list)
        self.widget_tree = None

    def get_interfaces(self):
        return [
            'app_shortcuts',
            'chkdeps',
            'gtk_shortcut_help',
            'gtk_window_listener',
            'screenshot_provider',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'screenshot',
                'defaults': ['openpaperwork_gtk.screenshots'],
            },
        ]

    def chkdeps(self, out: dict):
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)

    def on_gtk_window_opened(self, window):
        self.windows.append(window)

    def on_gtk_window_closed(self, window):
        self.windows.remove(window)

    def app_shortcut_add(
            self, shortcut_group, shortcut_desc, shortcut_keys, action_name):
        self.shortcut_help_add(
            shortcut_group, shortcut_desc, shortcut_keys, action_name
        )

    def shortcut_help_add(
            self, shortcut_group, shortcut_desc, shortcut_keys, action_name):
        LOGGER.info(
            "Keyboard shortcut: %s --> %s:%s",
            shortcut_keys, shortcut_group, shortcut_desc
        )
        group = self.groups[shortcut_group]
        group.append((
            shortcut_desc, shortcut_keys, action_name
        ))

    def gtk_show_shortcuts(self):
        LOGGER.info("Showing shortcuts")
        self.widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.shortcutswin", "shortcutswin.glade"
        )
        section = self.widget_tree.get_object("shortcuts_mainwindow")
        window = self.widget_tree.get_object("shortcuts")

        groups = {}
        for shortcut_group in sorted(list(self.groups.keys())):
            group = Gtk.ShortcutsGroup()
            group.set_property("title", shortcut_group)
            group.set_visible(True)
            groups[shortcut_group] = group
            section.add(group)

        for (shortcut_group, shortcuts) in self.groups.items():
            group = groups[shortcut_group]
            shortcuts = sorted(list(shortcuts))
            for (shortcut_desc, shortcut_keys, actions_name) in shortcuts:
                shortcut = Gtk.ShortcutsShortcut()
                shortcut.set_property("accelerator", shortcut_keys)
                shortcut.set_property("title", shortcut_desc)
                shortcut.set_visible(True)
                group.add(shortcut)

        window.set_transient_for(self.windows[-1])
        window.show_all()

    def gtk_hide_shortcuts(self):
        if self.widget_tree is None:
            return
        window = self.widget_tree.get_object("shortcuts")
        window.destroy()

    def screenshot_snap_all_doc_widgets(self, out_dir):
        if self.widget_tree is None:
            return
        self.core.call_success(
            "screenshot_snap_widget", self.widget_tree.get_object("shortcuts"),
            self.core.call_success("fs_join", out_dir, "shortcuts.png"),
        )



import logging

try:
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.promise
import openpaperwork_gtk.deps

from . import _


DELAY = 0.1
LOGGER = logging.getLogger(__name__)


class SurfacePreloader(object):
    def __init__(self, core, doc_id, doc_url, page_indexes):
        self.core = core
        self.doc_id = doc_id
        self.doc_url = doc_url
        self.page_indexes = page_indexes
        self.page_surfaces = []

    def start(self):
        LOGGER.info("Will load %d pages", len(self.page_indexes))
        for (page_nb, page_idx) in enumerate(self.page_indexes):
            page_url = self.core.call_success(
                "page_get_img_url", self.doc_url, page_idx
            )

            promise = openpaperwork_core.promise.Promise(
                self.core, self.core.call_all,
                args=(
                    "on_progress", "print_load",
                    page_nb / len(self.page_indexes),
                    _("Loading {doc_id} p{page_idx} for printing").format(
                        doc_id=self.doc_id, page_idx=(page_idx + 1)
                    )
                )
            )
            promise = promise.then(lambda *args, **kwargs: None)
            promise = promise.then(openpaperwork_core.promise.DelayPromise(
                self.core, DELAY
            ))
            promise = promise.then(self.core.call_success(
                "url_to_cairo_surface_promise", page_url
            ))
            promise = promise.then(self._add_surface)
            self.core.call_success("work_queue_add_promise", "print", promise)

        promise = openpaperwork_core.promise.Promise(
            self.core, self.core.call_all, args=(
                "on_progress", "print_load", 1.0
            )
        )
        self.core.call_success("work_queue_add_promise", "print", promise)

    def _add_surface(self, surface):
        LOGGER.info(
            "Got page rendering %d/%d",
            len(self.page_surfaces), len(self.page_indexes)
        )
        self.page_surfaces.append(surface)

    def get_nb_pages(self):
        return len(self.page_indexes)

    def get_nb_surfaces(self):
        return len(self.page_surfaces)

    def has_page_surface(self, page_nb):
        return page_nb < len(self.page_surfaces)

    def get_page_surface(self, page_nb):
        return self.page_surfaces[page_nb]

    def cancel(self):
        self.core.call_all("work_queue_cancel_all", "print")
        self.page_surfaces = []


class PrintJob(object):
    def __init__(
            self, core, preloader, window, doc_id, nb_pages, active_page_nb,
            job_name, default_filename):
        self.core = core
        self.preloader = preloader
        self.window = window
        self.doc_id = doc_id
        self.nb_pages = nb_pages

        print_settings = Gtk.PrintSettings()
        self.print_op = Gtk.PrintOperation()
        self.print_op.set_print_settings(print_settings)
        self.print_op.set_n_pages(nb_pages)
        if active_page_nb >= 0:
            self.print_op.set_current_page(active_page_nb)
        self.print_op.set_use_full_page(True)
        self.print_op.set_job_name(job_name)
        self.print_op.set_export_filename(default_filename)
        self.print_op.set_allow_async(True)
        self.print_op.set_embed_page_setup(True)
        self.print_op.set_show_progress(True)
        self.print_op.connect("status-changed", self._status_changed)
        self.print_op.connect("paginate", self._paginate)
        self.print_op.connect("preview", self._preview)
        self.print_op.connect("ready", self._preview_ready)
        self.print_op.connect("got_page_size", self._preview_got_page_size)
        self.print_op.connect("begin-print", self._begin)
        self.print_op.connect("request-page-setup", self._request_page_setup)
        self.print_op.connect("draw-page", self._draw)
        self.print_op.connect("end-print", self._end)
        self.print_op.connect("done", self._done)

    def _status_changed(self, print_op):
        LOGGER.info(
            "Print status: %s: %s",
            print_op.get_status(), print_op.get_status_string()
        )

    def _preview(self, print_op, print_preview, print_context, win_parent):
        LOGGER.info("User requested a preview")

    def _preview_got_page_size(self, print_preview, print_context, page_setup):
        LOGGER.info("Preview: got page size")

    def _preview_ready(self, print_preview, print_context):
        LOGGER.info("Preview ready")

    def _paginate(self, print_operation, print_context):
        pagination = self.preloader.get_nb_surfaces()
        running = (pagination < self.nb_pages)
        return not running

    def _begin(self, print_op, print_context):
        LOGGER.info("Printing has begun")
        self.core.call_all(
            "on_progress", "print", 0.0, _("Printing %s") % self.doc_id
        )

    def _request_page_setup(
            self, print_op, print_context, page_nb, page_setup):
        LOGGER.info(
            "Computing page setup for %d/%d",
            page_nb, self.nb_pages
        )

        surface = self.preloader.get_page_surface(page_nb)
        img_width = surface.surface.get_width()
        img_height = surface.surface.get_height()

        # take care of rotating the page if required
        img_portrait = (img_width <= img_height)
        LOGGER.info(
            "Page %d/%d: Orientation portrait = %s",
            page_nb, self.nb_pages, img_portrait
        )
        page_setup.set_orientation(
            Gtk.PageOrientation.PORTRAIT
            if img_portrait else
            Gtk.PageOrientation.LANDSCAPE
        )

    def _draw(self, print_op, print_context, page_nb):
        LOGGER.info(
            "Printing of %s %d/%d ; DPI: %fx%f",
            self.doc_id, page_nb, self.nb_pages,
            print_context.get_dpi_x(),
            print_context.get_dpi_y()
        )
        self.core.call_all(
            "on_progress", "print", page_nb / self.nb_pages,
            _("Printing {doc_id} ({page_idx}/{nb_pages})").format(
                doc_id=self.doc_id, page_idx=page_nb, nb_pages=self.nb_pages
            )
        )

        surface = self.preloader.get_page_surface(page_nb)
        img_width = surface.surface.get_width()
        img_height = surface.surface.get_height()

        scaling = min(
            print_context.get_width() / img_width,
            print_context.get_height() / img_height,
        )

        cairo_ctx = print_context.get_cairo_context()
        cairo_ctx.scale(scaling, scaling)
        cairo_ctx.set_source_surface(surface.surface)
        cairo_ctx.paint()

    def _end(self, print_op, print_context):
        self.refs = []
        LOGGER.info("Printing has ended")
        self.core.call_all("on_progress", "print", 1.0)

    def _done(self, print_op, print_op_result):
        LOGGER.info("Printing done")
        self.preloader.cancel()

    def run(self):
        self.print_op.run(Gtk.PrintOperationAction.PRINT_DIALOG, self.window)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.active_doc = None
        self.active_page_idx = None
        self.windows = []

        # WORKAROUND(Jflesch): keep a ref on the print operation to avoid
        # premature garbage collecting
        self._ref = None

    def get_interfaces(self):
        return [
            'chkdeps',
            'doc_open',
            'doc_print',
            'gtk_window_listener',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'cairo_url',
                'defaults': [
                    'paperwork_backend.cairo.pillow',
                    'paperwork_backend.cairo.poppler',
                ],
            },
            {
                'interface': 'work_queue',
                'defaults': ['openpaperwork_core.work_queue.default'],
            },
        ]

    def init(self, core):
        super().init(core)

        if not GTK_AVAILABLE:
            return

        self.core.call_success("work_queue_create", "print")

    def chkdeps(self, out: dict):
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)

    def on_gtk_window_opened(self, window):
        self.windows.append(window)

    def on_gtk_window_closed(self, window):
        self.windows.remove(window)

    def doc_open(self, doc_id, doc_url):
        self.active_doc = (doc_id, doc_url)

    def on_page_shown(self, page_idx):
        self.active_page_idx = page_idx

    def doc_print(self, doc_id, doc_url, page_indexes=None):
        active_page_idx = 0
        if self.active_doc[1] == doc_url:
            # prefer the current page when it makes sense
            active_page_idx = self.active_page_idx

        if page_indexes is None:
            nb_pages = self.core.call_success(
                "doc_get_nb_pages_by_url", doc_url
            )
            if nb_pages is None:
                raise Exception("No page in the document, Nothing to print")
            page_indexes = list(range(0, nb_pages))
            job_name = "Paperwork " + doc_id
            default_filename = doc_id + ".pdf"
        else:
            job_name = "Paperwork {} p{}".format(
                doc_id, ",".join((str(p) for p in page_indexes))
            )
            default_filename = "{}_p{}".format(
                doc_id,
                "_p".join((str(p) for p in page_indexes))
            )

        try:
            active_page_nb = page_indexes.index(active_page_idx)
        except ValueError:
            active_page_nb = -1

        preloader = SurfacePreloader(self.core, doc_id, doc_url, page_indexes)
        print_job = PrintJob(
            self.core, preloader, self.windows[-1], doc_id, len(page_indexes),
            active_page_nb, job_name, default_filename
        )

        LOGGER.info(
            "Opening print dialog for document %s (pages: %s)",
            doc_url, page_indexes
        )
        preloader.start()
        print_job.run()

        self._ref = print_job
